#include "dock.h"

/************************************************/
void Base_Score::calc_corner_coords() {
  float dx, dy, dz;

  dx = (span[0]-1)*spacing;
  dy = (span[1]-1)*spacing;
  dz = (span[2]-1)*spacing;

  x_min = origin[0];
  x_max = origin[0] + dx;
  y_min = origin[1];
  y_max = origin[1] + dy;
  z_min = origin[2];
  z_max = origin[2] + dz;

  corners[0].assign_vals(x_min, y_min, z_min);
  corners[1].assign_vals(x_max, y_min, z_min);
  corners[2].assign_vals(x_max, y_min, z_max);
  corners[3].assign_vals(x_min, y_min, z_max);
  corners[4].assign_vals(x_min, y_max, z_min);
  corners[5].assign_vals(x_max, y_max, z_min);
  corners[6].assign_vals(x_max, y_max, z_max);
  corners[7].assign_vals(x_min, y_max, z_max);

}

/************************************************/
bool Base_Score::check_box_boundaries(float x, float y, float z) {

  if( (x < x_min)||(x >= x_max)||(y < y_min)||(y >= y_max)||(z < z_min)||(z >= z_max) )
	return false;
  else
	return true;
}

/************************************************/
void Base_Score::find_grid_neighbors(float x, float y, float z) {
  int	x_below, x_above;
  int	y_below, y_above;
  int	z_below, z_above;
  int	x_nearest, y_nearest, z_nearest;
  float	x_int, y_int, z_int;
  float corrected_coords[3];

  x_int = x - origin[0];
  y_int = y - origin[1];
  z_int = z - origin[2];

  x_nearest = NINT(x_int/spacing);
  x_below = INTFLOOR(x_int/spacing);
//  x_below = (int)floor(x_int/spacing);
  x_above = x_below + 1;
  if(x_nearest >= span[0]) {
	if(x_below >= span[0])
		x_nearest = span[0] - 1;
	else
		x_nearest = x_below;
  }
  if(x_nearest < 0) {
	if(x_above < 0)
		x_nearest = 0;
	else
		x_nearest = x_above;
  }

  y_nearest = NINT(y_int/spacing);
  y_below = INTFLOOR(y_int/spacing);
//  y_below = (int)floor(y_int/spacing);
  y_above = y_below + 1;
  if(y_nearest >= span[1]) {
	if(y_below >= span[1])
		y_nearest = span[1] - 1;
	else
		y_nearest = y_below;
  }
  if(y_nearest < 0) {
	if(y_above < 0)
		y_nearest = 0;
	else
		y_nearest = y_above;
  }

  z_nearest = NINT(z_int/spacing);
  z_below = INTFLOOR(z_int/spacing);
//  z_below = (int)floor(z_int/spacing);
  z_above = z_below + 1;
  if(z_nearest >= span[2]) {
	if(z_below >= span[2])
		z_nearest = span[2] - 1;
	else
		z_nearest = z_below;
  }
  if(z_nearest < 0) {
	if(z_above < 0)
		z_nearest = 0;
	else
		z_nearest = z_above;
  }

  nearest_neighbor = find_grid_index(x_nearest, y_nearest, z_nearest);

  neighbors[0] = find_grid_index(x_above, y_above, z_above);
  neighbors[1] = find_grid_index(x_above, y_above, z_below);
  neighbors[2] = find_grid_index(x_above, y_below, z_above);
  neighbors[3] = find_grid_index(x_below, y_above, z_above);
  neighbors[4] = find_grid_index(x_above, y_below, z_below);
  neighbors[5] = find_grid_index(x_below, y_above, z_below);
  neighbors[6] = find_grid_index(x_below, y_below, z_above);
  neighbors[7] = find_grid_index(x_below, y_below, z_below);

  corrected_coords[0] = x - origin[0];
  corrected_coords[1] = y - origin[1];
  corrected_coords[2] = z - origin[2];

  int_coords[0] = corrected_coords[0];
  int_coords[1] = corrected_coords[1];
  int_coords[2] = corrected_coords[2];

  cube_coords[0] = corrected_coords[0]/spacing - (float)(INTFLOOR(corrected_coords[0]/spacing)); // removed (int)
  cube_coords[1] = corrected_coords[1]/spacing - (float)(INTFLOOR(corrected_coords[1]/spacing)); // removed (int)
  cube_coords[2] = corrected_coords[2]/spacing - (float)(INTFLOOR(corrected_coords[2]/spacing)); // removed (int)
}

/************************************************/
int Base_Score::find_grid_index(int x, int y, int z) {

  return span[0] * span[1] * z + span[0] * y + x;

}

/************************************************/
float Base_Score::interpolate(float *grid) {
  float a1,a2,a3,a4,a5,a6,a7,a8;
  float value;
  int	out_of_bounds, i;

  out_of_bounds = 0;
  for(i=0;i<8;i++)
	  if((neighbors[i] > size)||(neighbors[i] < 0))
		  out_of_bounds = 1;

  if(out_of_bounds == 0) {
    a8 = grid[neighbors[7]];
    a7 = grid[neighbors[6]] - a8;
    a6 = grid[neighbors[5]] - a8;
    a5 = grid[neighbors[4]] - a8;
    a4 = grid[neighbors[3]] - a8 - a7 - a6;
    a3 = grid[neighbors[2]] - a8 - a7 - a5;
    a2 = grid[neighbors[1]] - a8 - a6 - a5;
    a1 = grid[neighbors[0]] - a8 - a7 - a6 - a5 - a4 - a3 - a2;

    value = a1*cube_coords[0]*cube_coords[1]*cube_coords[2] + a2*cube_coords[0]*cube_coords[1] +
	        a3*cube_coords[0]*cube_coords[2] + a4*cube_coords[1]*cube_coords[2] +
	  	    a5*cube_coords[0] + a6*cube_coords[1] + a7*cube_coords[2] + a8;

    return value;
  } else {
	return grid[nearest_neighbor];
  }

}

