#include "dock.h"

  int					ie_att_exp;
  int 					ie_rep_exp;
  float					ie_diel;
  int					max_conf;


/************************************************/
int main(int argc, char **argv) {

  // DOCK Classes
  // DOCK does not use constructors, but each class has an initialize member.
  // Note that the compiler-generated default constructors effectively
  // do nothing; so that these objects should be considered uninitialized
  // until their initialize members are called below.
  Parameter_Reader			c_parm;
  Library_File				c_library;
  Orient					c_orient;
  Master_Conformer_Search	c_master_conf;
  Bump_Filter				c_bmp_score;
  Simplex_Minimizer			c_simplex;
  Master_Score				c_master_score;
  AMBER_TYPER				c_typer;
  DOCKMol					mol;
  DOCKMol					mol_ref_ext;
  vector<DOCKMol> 				mol_conf_printed;
  

  // Local vars
  int						i, length;
  ofstream					outfile;

  char						fname[50];
  bool						USE_MPI;
  int						max_conf;
  char						klient[1000];  

  // Preprocessor macro MPI is controlled by the platform configuration
  // which is specified during installation.
#ifdef BUILD_DOCK_WITH_MPI
  	USE_MPI = true;
#else
	USE_MPI = false;
#endif

  //initialize mpi and determine if proper number of processors has been called
  if (USE_MPI)
      	// mpi_init requires pointers to argc and argv.
      	USE_MPI = c_library.initialize_mpi( & argc, & argv);

  /////////////////// Direct the output to a file or stdout ////////////////////
  if(check_commandline_argument(argv, argc, "-o") != -1) {
  	if(USE_MPI) {
		  if(c_library.rank > 0) {
			sprintf(fname, "%s.%d", parse_commandline_argument(argv, argc, "-o").c_str(), c_library.rank);
		  	outfile.open(fname);
		 	cout.rdbuf(outfile.rdbuf());
	  	  } else {
		  	sprintf(fname, "%s", parse_commandline_argument(argv, argc, "-o").c_str());
		  	outfile.open(fname);
		  	cout.rdbuf(outfile.rdbuf());
	  	  }

  	} else {
	  	  sprintf(fname, "%s", parse_commandline_argument(argv, argc, "-o").c_str());
	  	  outfile.open(fname);
	  	  cout.rdbuf(outfile.rdbuf());
        }

  } else if(USE_MPI) {
	  cout << "Error: DOCK must be run with the -o outfile option under MPI" << endl;
  	  c_library.finalize_mpi();
	  exit(0);
  }
  ////////////////////////////////////////////////////////////////////////////
  // Begin timing
  double start_time = wall_clock_seconds();
  
  print_header();

  // Read input parameters
  c_parm.initialize(argc, argv);
  c_library.input_parameters(c_parm, USE_MPI, max_conf);
  c_orient.input_parameters(c_parm);
  c_master_conf.input_parameters(c_parm);
  c_bmp_score.input_parameters(c_parm);
  c_master_score.input_parameters(c_parm);
  c_simplex.input_parameters(c_parm);
  mol.input_parameters(c_parm);
  c_typer.input_parameters(c_parm);
  if (!USE_MPI)
  	  c_parm.write_params();
  
  // Exit the program if parameterization fails //
  if(!c_parm.parameter_input_successful()) {
	if(USE_MPI)
  		c_library.finalize_mpi();
	if(outfile.is_open())
		outfile.close();
	return 0;
  }
  ////////////////////////////////////////////////

  // Initialization routines
  c_typer.initialize();
  c_library.initialize(argc, argv, USE_MPI);
  c_orient.initialize(argc, argv);
  c_master_conf.initialize();
  c_bmp_score.initialize();
  c_master_score.initialize_all(c_typer, argc, argv);
  c_simplex.initialize();
  c_library.submit_rmsd_reference_file_open();
  
#ifdef BUILD_DOCK_WITH_MPI
  if((USE_MPI) && (c_library.rank > 0)){
	  cout << "DOCK is currently running on ";
	  MPI_Get_processor_name(klient, &length);
	  for(i=0;i<length;i++)
	  	cout << klient[i];
  	  cout << endl;
  }	  
#endif


  // Main loop
  while(c_library.get_mol(mol, USE_MPI)) {
   
	  mol.prepare_molecule();
	  c_typer.prepare_molecule(mol);
	  c_master_conf.prepare_molecule(mol);
	  mol_conf_printed.clear();

          c_library.submit_conformation(mol, mol_ref_ext, mol_conf_printed, c_master_score, c_master_conf, max_conf);

	  while(c_master_conf.next_anchor(mol)) {		  

		  c_library.num_anchors++;

		  c_orient.match_ligand(mol);

		  while(c_orient.new_next_orientation(mol)) {

			  if(c_bmp_score.check_bumps(mol, c_orient.more_orientations())) {

				  c_simplex.minimize_rigid_anchor(mol, c_master_score);
				  
				  
				  c_library.submit_orientation(mol, c_master_score);
				  
				  if(c_master_conf.submit_anchor_orientation(mol, c_orient.more_orientations(), c_master_score)) {

					    c_master_conf.grow_periphery(c_master_score, c_bmp_score, c_simplex);

					    while(c_master_conf.next_conformer(mol)) {

						  c_simplex.minimize_final_pose(mol, c_master_score);
	
						  c_library.submit_conformation(mol, mol_ref_ext, mol_conf_printed, c_master_score, c_master_conf, max_conf);

						  c_library.submit_scored_pose(mol, c_master_score);
//  added MMiteva						  
						  if( mol_conf_printed.size() == max_conf)
								goto Printconf_end;
//  end MMiteva						  	

						  
					    } // End Grow Periphery Loop
					  
				  }  // End Submit Anchor Orientation Loop

			  } // End Bump Check

		  } // End Orientation Loop

	  } // End Anchor Loop
	 
Printconf_end:
	cout << endl << " Print_mol End" << endl;

	  
  } // End Ligand Loop


  c_library.submit_rmsd_reference_file_close();

  if((!USE_MPI) || (c_library.rank == 0))
  	cout << endl << endl << (c_library.total_mols - c_library.initial_skip) << " Molecules Processed" << endl;
  else
	cout << endl << endl << c_library.completed << " Molecules Processed" << endl;

  // End timing
  double stop_time = wall_clock_seconds();
  cout << "Elapsed time:\t" << stop_time - start_time << " seconds" << endl;

  if(USE_MPI)
 	c_library.finalize_mpi();

  if(outfile.is_open())
	  outfile.close();
	  

  return 0;

}

/************************************************/
double wall_clock_seconds() {
  time_t t;
  if ( static_cast<time_t>( -1 ) == time( &t) ) {
	  cout << "Error from time function!  Elapsed time is erroneous." << endl;
  }
  return static_cast<double>( t );
}
/************************************************/
void print_header() {

	cout << endl << endl << endl;
	cout << "--------------------------------------" << endl;
	cout << "Generation of 3D multiple conformations by Multiconf-DOCK" << endl << endl;
	cout << "--------------------------------------" << endl << endl;
}
