#include "dock.h"
#include <sstream>

/************************************************/
string Library_File::hdb_atom_type_converter(int numeric_type) {
	string	type;

	switch (numeric_type) {

		case 1:
			type = "C.";
			break;

		case 5:
			type = "C.3";
			break;

		case 6:
			type = "H";
			break;

		case 7:
			type = "H";
			break;

		case 8:
			type = "N.";
			break;

		case 9:
			type = "N.4";
			break;

		case 10:
			type = "N.3";
			break;

		case 11:
			type = "O.";
			break;

		case 12:
			type = "O.3";
			break;

		case 13:
			type = "P."; // possible degeneracy with P. or P2
			break;

		case 14:
			type = "S.";
			break;

		case 15:
			type = "F";
			break;

		case 16:
			type = "Cl";
			break;

		case 17:
			type = "Br";
			break;

		case 18:
			type = "I";
			break;

		case 19:
			type = "Na"; // possible degeneracy with Na or K
			break;

		case 20:
			type = "Li"; // possible degeneracy with Li, Al or B
			break;

		case 21:
			type = "Ca";
			break;

		case 24:
			type = "Si";
			break;

		case 25:
			type = "Du";
			break;

		default:
			type = "Du";
			cout << "Atom type conversion not found for atom type " << numeric_type << ".  Du type assumed." << endl;
			break;
	}

	return type;

}

/************************************************/
//bool Library_File::read_hierarchy_db(vector<DOCKMol> &mol_branches, ifstream &ifs) {
  bool Library_File::read_hierarchy_db(DOCKMol &mol, ifstream &ifs) {
	int				i, j, k, l;
	char			line[1000];
	string			tmp;
	int				total_atoms;
	vector<DOCKMol>	mol_branches;

	// family level data
	int				clunum, nmol, nbr, fbr;

	// molecule level data
	string			name, refcode;

	// branch level data
	int				beatm, bmatm, bnhvy, bnhyd, confnum, bi, iconf;
	float			solvat, apol;

	// new molecule data
	DOCKMol			new_mol;
	FLOATVec		charges;
	vector<string>	types;
	vector<string>	summary_lines;
	string			family_line, name_line, branch_line;
	INTVec			level_counts, level_values, level_offsets;
	int				current_level, current_count, index;
	INTVec			atom_levels;
	int				next_level, last_level;

	mol_branches.clear();

	// if file has no lines, return false
	if(!ifs.getline(line, 1000))
		return false;

	do {

		if(!strncmp(line, "Family", 6)) {

			// read vars from Family line
			sscanf(line, "%*s %d %d %d %d", &clunum, &nmol, &nbr, &fbr);
			family_line = line;

			nbr++;
			fbr++;

			// loop over nmols
			for(i=0;i<nmol;i++) {

				//////////////////////
				// skip first MBR loop
				//////////////////////

				// read in mol name & refcode
				if(!ifs.getline(line, 1000)) {
					cout << "Error: Incomplete hierarchy molecule record." << endl;
					return false;
				}

				name_line = line;
				name = name_line.substr(0,47);
				refcode = name_line.substr(47, 56);

				///////////////////////
				// skip the other MBR variables
				///////////////////////

				// loop over branches
				for(j=0;j<fbr;j++) {

					// read branch header line
					if(!ifs.getline(line, 1000)) {
						cout << "Error: Incomplete hierarchy molecule record." << endl;
						return false;
					}

					sscanf(line, "%d %d %d %d %f %d %f %d %d", &beatm, &bmatm, &bnhvy, &bnhyd, &solvat, &confnum, &apol, &bi, &iconf);
					branch_line = line;

					// create new DOCKMol to store each branch
					new_mol.clear_molecule();
					new_mol.allocate_arrays(beatm, 0);
					new_mol.title = name;

					// assign heirarchy data to mol_data string
					if(j==0) {
						new_mol.mol_data  = family_line + "\n";
						new_mol.mol_data += name_line + "\n";
					}
					new_mol.mol_data += branch_line;

					// allocate heirarchy data structures
					charges.resize(bmatm);
					types.resize(bmatm);
					summary_lines.resize(bmatm);

					level_counts.clear();
					level_counts.push_back(0);
					level_values.clear();
					level_values.push_back(0);

					atom_levels.resize(beatm, 0);

					current_level = 0;
					next_level = 0;
					last_level = 0;

					// read the info into infovec struct
					for(k=0;k<bmatm;k++) {
						ifs.getline(line, 1000);
						tmp = line;

						types[k] = hdb_atom_type_converter(atoi(tmp.substr(3,2).c_str()));
						charges[k] = atof(tmp.substr(5,5).c_str())*0.001;

						summary_lines[k] = "";
						summary_lines[k] += tmp.substr(0,3) + "\n";  // level
						summary_lines[k] += tmp.substr(3,2) + "\n"; // vdwtype
						sprintf(line, "%f\n", charges[k]);
						summary_lines[k] += line; // charge
						summary_lines[k] += tmp.substr(10,1) + "\n"; // flagat
						summary_lines[k] += tmp.substr(11,2) + "\n"; // lcolor
						summary_lines[k] += tmp.substr(13,9) + "\n"; // polsolv
						summary_lines[k] += tmp.substr(22,9) + "\n"; // apolsolv
						summary_lines[k] += types[k] + "\n"; // converted type

						next_level = atoi(tmp.substr(0,3).c_str());

						// count the number of atoms in each level
						if(k>0) {
							if(next_level != last_level) { // !=, >
								level_counts.push_back(0);
								level_values.push_back(0);
								current_level++;
								last_level = next_level;
							}
						} else
							last_level = next_level;

						level_values[current_level] = next_level;
						level_counts[current_level]++;
					}

					// calculate the level offsets
					level_offsets.clear();
					level_offsets.resize(level_counts.size(),0);

					for(k=1;k<level_offsets.size();k++) {
						level_offsets[k] = level_offsets[k-1] + level_counts[k-1];
					}

					// add the level count information to mol_data
					new_mol.mol_data += "\n";

					for(k=0;k<level_counts.size();k++) {
						sprintf(line, "%d %d,", level_counts[k], level_values[k]);
						new_mol.mol_data += line;
					}

					current_level = 0;
					current_count = 0;

					// read coords for all confs
					for(k=0;k<beatm;k++) {

						// read one line
						ifs.getline(line, 1000);
						tmp = line;

						// assign coordinates
						new_mol.x[k] = atoi(tmp.substr(3,6).c_str())*0.001;
						new_mol.y[k] = atoi(tmp.substr(9,6).c_str())*0.001;
						new_mol.z[k] = atoi(tmp.substr(15,6).c_str())*0.001;

						atom_levels[k] = atoi(tmp.substr(0,3).c_str());

//						cout << atom_levels[k] << "\t";//////////////////////////////

						// scan for level transitions
						if(k>0) {

							// if level has increased
							if(atom_levels[k] > atom_levels[k-1]) {

//								cout << ".";////////////////////////////////////////

								current_level = 0;

								for(l=0;l<level_values.size();l++) {

//									cout << level_values[l] << " ";//////////////

									if(atom_levels[k] == level_values[l]) {
										current_level = l;
										break;
									}
								}

								current_count = 0;
							}

							// if level has decreased
							if(atom_levels[k] < atom_levels[k-1]) {
								current_level = 0;

								for(l=0;l<level_values.size();l++) {
									if(atom_levels[k] == level_values[l]) {
										current_level = l;
										break;
									}
								}

								current_count = 0;
							}

							if(current_count >= level_counts[current_level])
								current_count = 0;
						}

//						cout << current_level << endl;/////////////////////////////////

						// apply vdw & charge values from info section to atom records
						index = level_offsets[current_level] + current_count;

						new_mol.charges[k] = charges[index];
						new_mol.atom_data[k] = summary_lines[index];
						new_mol.atom_types[k] = types[index];
						new_mol.atom_names[k] = types[index];

						current_count++;

					}

					// assign branches of molecule to vector of mols
					mol_branches.push_back(new_mol);

				}

			}

			// merge molecules into one total mol

			// count all the atoms
			total_atoms = 0;
			for(i=0;i<mol_branches.size();i++)
				total_atoms += mol_branches[i].num_atoms;

			// create new hybrid molecule with all branches
			mol.clear_molecule();
			mol.allocate_arrays(total_atoms, 0);

			index = 0;

			for(i=0;i<mol_branches.size();i++) {

				for(j=0;j<mol_branches[i].num_atoms;j++) {

					// copy molecule atom data
					mol.x[index] = mol_branches[i].x[j];
					mol.y[index] = mol_branches[i].y[j];
					mol.z[index] = mol_branches[i].z[j];

					mol.charges[index]    = mol_branches[i].charges[j];
					mol.atom_types[index] = mol_branches[i].atom_types[j];
					mol.atom_names[index] = mol_branches[i].atom_names[j];

					mol.atom_data[index]  = mol_branches[i].atom_data[j];

					// no bonds, hence no rings
					mol.atom_ring_flags[index] = false;

					// activate atoms
					mol.atom_active_flags[index] = true;///////////////

					index++;

				}

				mol.mol_data += mol_branches[i].mol_data;
				mol.mol_data += "\n";

			}

			mol.num_active_atoms = total_atoms;

			mol_branches.clear();
			// return true if mol is found
			return true;
		}

	} while (ifs.getline(line, 1000));

	// if no mol found, return false
	return false;

}

/************************************************/
void Library_File::input_parameters(Parameter_Reader &parm, bool USE_MPI, int &max_conf) {
  string tmp;

  cout << "\nMolecule Library Parameters" << endl;
  cout << "------------------------------------------------------------------------------------------" << endl;

  input_file_name = parm.query_param("ligand_atom_file","database.mol2");
  output_file_prefix = parm.query_param("ligand_outfile_prefix","output");


  max_mol_limit = (parm.query_param("limit_max_ligands", "no", "yes no") == "yes") ? true : false;
  if(max_mol_limit)
	  max_mols = atoi(parm.query_param("max_ligands", "1000").c_str());
  else
	  max_mols = -1;

//  tmp = parm.query_param("write_orientations", "no", "yes no"); Com. by MMiteva
  tmp = "no";
  	
  if(tmp == "yes")
	  if(!USE_MPI)
	  	write_orients = true;
  	  else{
		  cout << "Error:  DOCK cannot write orientations while running in parallel." << endl;
		  finalize_mpi();
		  exit(0);
	  }
  else
	  write_orients = false;

  tmp = parm.query_param("write_conformations", "no", "yes no");
  if(tmp == "yes")
	  if(!USE_MPI)
	  	write_conformers = true;
  	  else{
		  cout << "Error:  DOCK cannot write conformations while running in parallel." << endl;
		  finalize_mpi();
		  exit(0);
	  }
  else
	  write_conformers = false;

  initial_skip = atoi(parm.query_param("initial_skip", "0").c_str());
  
  max_conf = atoi(parm.query_param("max_conformations", "50").c_str());
  

  tmp = parm.query_param("calculate_rmsd", "no", "yes no");
  if(tmp == "yes")
	  calc_rmsd = true;
  else
	  calc_rmsd = false;

  if(calc_rmsd) {
	  constant_rmsd_ref = (parm.query_param("use_rmsd_reference_mol", "no") == "yes") ? true : false;
	  if(constant_rmsd_ref) {
		  constant_rmsd_ref_file = parm.query_param("rmsd_reference_filename", "ligand_rmsd.mol2");
	  }
  } else {
	  constant_rmsd_ref = false;

  }

  rmsd_window = atof(parm.query_param("rmsd_window", "1.0").c_str());

  energy_window = atof(parm.query_param("energy_window", "100.0").c_str());
//  tmp = parm.query_param("rank_ligands", "no", "yes no"); //Com. by MMiteva
  tmp = "no";
  if(tmp == "yes")
	  rank_ligands = true;
  else
	  rank_ligands = false;

  if(rank_ligands) {
	  max_ranked = atoi(parm.query_param("max_ranked_ligands", "500").c_str());

	  tmp = parm.query_param("scored_mol_output_override", "no", "yes no");
	  if(tmp == "yes")
		  rank_ligand_output_override = true;
	  else
		  rank_ligand_output_override = false;
 	  num_scored_poses = max_ranked;
 	  cluster_ranked_poses = (parm.query_param("cluster_ranked_poses", "yes", "yes no") == "yes") ? true : false;                                     
   	  if(cluster_ranked_poses)                                                      
 		  cluster_rmsd_threshold = atof(parm.query_param("cluster_rmsd_threshold", "2.0").c_str());                                                                        
  }
  if(!rank_ligands){
//	  num_scored_poses = atoi(parm.query_param("num_scored_poses_written", "1").c_str());//Com.by MMiteva
	  num_scored_poses = 1;
	  cluster_ranked_poses = false;
  }


}

/************************************************/
void Library_File::initialize(int argc, char **argv, bool USE_MPI) {

  cout << endl;
  cout << "Initializing Library File Routines...\n";
  total_mols = 0;
  completed = 0;

  output_file_orient = output_file_prefix + "_orients.mol2";
  output_file_confs  = output_file_prefix + "_conformers.mol2";
  output_file_scored = output_file_prefix + "_scored.mol2";
  output_file_ranked = output_file_prefix + "_ranked.mol2";

  bestmol.clear_molecule();
  bestscore = 0.0;

  if(rank_ligands)
	  ranked_list.reserve(max_ranked);

  if(rank_ligands)
	  write_scored_ligands = rank_ligand_output_override;
  else
	  write_scored_ligands = true;

  if ((!USE_MPI) || (is_master_node()))
	  open_files();

  rank_list_highest_energy = 0.0;
  rank_list_highest_pos = -1;

  read_success = false;

}

/************************************************/
void Library_File::open_files() {
  DOCKMol	tmp_mol;
  int		i;
  string	extension;

  // identify the input file type: MOL2 or HDB (if not clear- assume MOL2)
  extension = input_file_name.substr(input_file_name.rfind("."), input_file_name.size());

  if(extension == ".db") {
	  db_type = 2;
  } else if (extension == ".mol2") {
	  db_type = 1;
  } else {
	  cout << "Warning: Filetype not recognized for " << input_file_name << ".  Assuming MOL2 format" << endl;
	  db_type = 1;
  }

  ligand_in.open(input_file_name.c_str());
  if(ligand_in.fail()) {
	  cout << "\n\nCould not open " << input_file_name << " for reading.  Program will terminate." << endl << endl;
	  exit(0);
  }

  if(write_orients) {
	  ligand_out_orients.open(output_file_orient.c_str());
	  if(ligand_out_orients.fail()) {
		  cout << "\n\nCould not open " << output_file_orient << " for writing.  Program will terminate." << endl << endl;
		  exit(0);
	  }
  }

  if(write_conformers) {
	  ligand_out_confs.open(output_file_confs.c_str());
	  if(ligand_out_confs.fail()) {
		  cout << "\n\nCould not open " << output_file_confs << " for writing.  Program will terminate." << endl << endl;
		  exit(0);
	  }
  }

//comm. by MMiteva
//  if(write_scored_ligands) {
//	  ligand_out_scored.open(output_file_scored.c_str());
//	  if(ligand_out_scored.fail()) {
//		  cout << "\n\nCould not open " << output_file_scored << " for writing.  Program will terminate." << endl << endl;
//		  exit(0);
//	  }
// }


  if(rank_ligands) {
	  ligand_out_ranked.open(output_file_ranked.c_str());
	  if(ligand_out_ranked.fail()) {
		  cout << "\n\nCould not open " << output_file_ranked << " for writing.  Program will terminate." << endl << endl;
		  exit(0);
	  }
  }

  // read through the initial skip ligands
  for(i=0;i<initial_skip;i++) {
	  read_mol(tmp_mol);
  }

  if(initial_skip > 0)
	  cout << " Skipped " << total_mols << " ligands" << endl << endl;

}

/************************************************/
void Library_File::close_files() {

  ligand_in.close();

  ligand_out_orients.close();
  ligand_out_confs.close();
  ligand_out_scored.close();
  ligand_out_ranked.close();

}

/************************************************/
bool Library_File::read_mol(DOCKMol &mol) {

  mol.clear_molecule();

  if(db_type == 1) { // if file is a MOL2 file

	  if(max_mol_limit && (total_mols >= max_mols))
		  return false;
	  
	  if(Read_Mol2(mol, ligand_in)) {
		  total_mols++;	  
		  return true;
	  } else{ 
		  return false;
	  }

  } else if(db_type == 2) { // if file is a hierarchy db
/**
	  // read a new mol if branch list is empty
	  if(mol_branches.size() == 0) {
		  if(!read_hierarchy_db(mol_branches, ligand_in))
			  return false;
	  }

	  // take one branch and copy it to mol
	  copy_molecule(mol, mol_branches[mol_branches.size()-1]);
	  mol_branches.pop_back();

	  // if last branch, count the molecule as read
	  if(mol_branches.size() == 0)
		  total_mols++;
**/

	  if(read_hierarchy_db(mol, ligand_in)) {
		  total_mols++;

		  // fail if too many mols have been read
		  if(max_mol_limit && (total_mols > max_mols))
			  return false;
		  else
			  return true;

	  } else {
		  return false;
	  }

  } else {

	  return false;

  }

}

/************************************************/
void Library_File::write_mol(DOCKMol &mol, ofstream &ofs) {

  Write_Mol2(mol, ofs);

}

/************************************************/
bool Library_File::get_mol(DOCKMol &mol, bool USE_MPI) {
	RANKMol	tmp_mol;
	int		i;

   if(!USE_MPI) { // single proc code
	  if(read_success){
		if(ranked_poses.size() > 0) {
			write_scored_poses();
		}else{
			cout << endl << "-----------------------------------" << endl;
			cout << "ERROR:  Could not find valid orientation!" << endl;
		}
	  }

	  // pose ranking code
	  ranked_poses.clear();
	  high_position = 0;
	  high_score = 0.0;

	  if(read_mol(mol)) {

		  num_orients = 0;
		  num_confs = 0;
		  num_anchors = 0;
		  printed_conf = 0;

		  if(calc_rmsd)
			  submit_rmsd_reference(mol);

		  read_success = true;

		  return true;
	  } else {

		  if(rank_ligands)
			  write_ranked_ligands();

		  close_files();
		  return false;
	  }

  } else {
/***/
	  if(is_master_node()) { // master node code

		  while(listen_for_request()) {

			  if(is_send_request()) {

				  while(continue_reading_mols()) {

					  if(read_mol(mol)) {
						  add_to_send_queue(mol);

						  // if HDB mols in send queue, then set continue mols true to allow all to be sent
						  if(mol_branches.size() > 0)
							  continue_mols = true;

						  read_success = true;
					  } else {
						  continue_mols = false;
					  }

				  } // End loop over work unit of mols

				  transmit_send_queue();

			  } // End send mol code

			  if(is_receive_request()) {

				  receive_flag = true;

				  ranked_poses.clear();

				  while(get_from_recv_queue(tmp_mol)) {

					  ranked_poses.push_back(tmp_mol);

				  } // End loop over received mols

				  // copy docking stats from MPI stats struct
				  num_anchors = stats[0];
				  num_orients = stats[1];
				  num_confs = stats[2];
				  
				  if(read_success){
						if(ranked_poses.size() > 0) {
							write_scored_poses();
						}else{
							cout << endl << "-----------------------------------" << endl;
							cout << "ERROR:  Could not find valid orientation!" << endl;
						}
				  }

			  } // End receive mol code

		  } // End looping over MPI listener

		  if(rank_ligands)
			  write_ranked_ligands();

		  close_files();
		  return false;

	  } else if(is_client_node()) { // client node code

		  // sort ranked poses
		  if(ranked_poses.size() > 0){
			  completed++;
			  sort(ranked_poses.begin(), ranked_poses.end());
		  }

		  // add ranked poses to return list
		  recv_queue.clear();
		  for(i=0;i<ranked_poses.size();i++)
			  add_to_recv_queue(ranked_poses[i]);

		  // copy docking stats to MPI data struct
		  stats[0] = num_anchors;
		  stats[1] = num_orients;
		  stats[2] = num_confs;

		  // init pose ranking
		  ranked_poses.clear();
		  high_position = 0;
		  high_score = 0.0;

		  // get molecule from server
		  if(get_from_send_queue(mol)) {

			  num_anchors = 0;
			  num_orients = 0;
			  num_confs = 0;

			  if(calc_rmsd)
				  submit_rmsd_reference(mol);
			  
			  read_success = true;
			  
			  return true;
		  } else return false;

	  }
/***/
	  return false; // to squelch warnings
  }

}

/************************************************/
void Library_File::submit_orientation(DOCKMol &mol, Master_Score &score) {

  num_orients++;
  
  if(write_orients) {
		  ligand_out_orients << "\n########## Name:\t\t" << mol.title << "_" << num_orients << endl;

		  if(calc_rmsd)
			  ligand_out_orients << "########## RMSD:\t\t" << calculate_rmsd(mol) << endl;

		  ligand_out_orients << endl;

		  write_mol(mol, ligand_out_orients);

  }

}
#define Abs(x) ((x) < 0 ? -(x) : (x))
/************************************************/
void Library_File::submit_conformation(DOCKMol &mol, DOCKMol &mol_ref_ext, vector<DOCKMol> &mol_conf_printed, Master_Score &score, Master_Conformer_Search &master,
int &max_conf) {
  float intra_en, intra_rmsd, diff_en;
  int	icheck;
  string original_mol_title;

       
    if(mol_conf_printed.size() == 0)
  {
  		 		 	
	         if(constant_rmsd_ref) 
			Read_Mol2(mol_ref_ext, rmsd_ref);
  
  		  printed_conf++;
		  submit_rmsd_reference(mol);
	    	  ref_en = calculate_Intra_Lig_Energy(mol, master);
		  ligand_out_confs << "\n########## Name:\t\t" << mol.title << "_" << printed_conf << endl;			
		  ligand_out_confs << "########## RMSD_initial:\t\t" << calculate_rmsd(mol) << endl;	
		  
   	  	    if(constant_rmsd_ref) {
		  	submit_rmsd_reference(mol_ref_ext);
		  	ligand_out_confs << "########## RMSD_external:\t\t" << calculate_rmsd(mol) << endl;
		  	submit_rmsd_reference(mol);	
		    } 
		  
		  		 
		  ligand_out_confs << "########## Internal Ligand Energy (L-J + Coulomb):\t\t" << ref_en << endl;

//-------------new to add num_conf on the title--------------------------------
		  std:: ostringstream oss;
		  oss << printed_conf;
		  std::string resultt = oss.str();
		  original_mol_title =  mol.title;	
		  mol.title = mol.title + '_' + resultt;
//new end----------------------------------------------------/

 		  mol_conf_printed.push_back(mol);
  		  write_mol(mol, ligand_out_confs);
		  mol.title = original_mol_title;
  }
  
  else
  
  {
   num_confs++;	
//   if(write_conformers) {

	 if(printed_conf < max_conf)  {
                  		  
	  intra_en = calculate_Intra_Lig_Energy(mol, master);
	  diff_en = Abs( ref_en - intra_en );	  
	  if((intra_en < energy_window) || (diff_en < energy_window)) {
	  
		intra_rmsd = calculate_rmsd(mol);  
		if(intra_rmsd > rmsd_window)	{

		 icheck = 1;	
  		 for( int i=0; i < mol_conf_printed.size(); i++ ) {
		   submit_rmsd_reference(mol_conf_printed[i]);
		   if( calculate_rmsd(mol) < rmsd_window) {
		   	icheck = 0;
//			cout << "\n icheck=0,  TEST RMSD =  " << calculate_rmsd(mol) << endl;	//test write	
		   }
 		 }
		 if(icheck == 1) {
		 submit_rmsd_reference(mol_conf_printed[0]);
		 		
		  printed_conf++;
//		  ligand_out_confs << "\n########## Name:\t\t" << mol.title << "_" << num_confs << endl;

		  ligand_out_confs << "\n########## Name:\t\t" << mol.title << "_" << printed_conf << endl;
		
		  ligand_out_confs << "########## RMSD_initial:\t\t" << calculate_rmsd(mol) << endl;
		  ligand_out_confs << "########## RMSD_intra:\t\t" << intra_rmsd << endl;
		  
 	  	  if(constant_rmsd_ref) {
		  	submit_rmsd_reference(mol_ref_ext);
			
		  	ligand_out_confs << "########## RMSD_external:\t\t" << calculate_rmsd(mol) << endl;			
		  } 
		  	 
		  ligand_out_confs << "########## Internal Ligand Energy (L-J + Coulomb):\t\t" << intra_en << endl;


//-------------new to add num_conf on the title--------------------------------
		  std:: ostringstream oss;
		  oss << printed_conf;
		  std::string resultt = oss.str();
		  original_mol_title =  mol.title;	
		  mol.title = mol.title + '_' + resultt;
//new end----------------------------------------------------/

 		  mol_conf_printed.push_back(mol);
		  submit_rmsd_reference(mol);
		  write_mol(mol, ligand_out_confs);
		  mol.title = original_mol_title;
		  
//Important test for rmsd:		  
//		  write_mol(rmsd_reference, ligand_out_confs); 

		}
	       }
	     }		
	}

  }
   
  
}

/************************************************/
void Library_File::submit_scored_pose(DOCKMol &mol, Master_Score &score) {
  char		score_convert[100];
  RANKMol	tmp_rankmol;
  int		i;
  float		max_score;
  int		max_position;
  bool		success;

  if(score.use_score) {
	  success = score.compute_secondary_score(mol);
	  if(success){
		  mol.current_data = score.secondary_score->output_score_summary(mol.current_score);
	  }
  } else {
	  mol.current_score = 0.0;
	  mol.current_data = "";
  }

  // Multi-pose output
  if(ranked_poses.size() < num_scored_poses) {
	  // Add the molecule
	  ranked_poses.push_back(tmp_rankmol);
	  ranked_poses[ranked_poses.size()-1].score = mol.current_score;
	  ranked_poses[ranked_poses.size()-1].data = mol.current_data;
	  copy_molecule(ranked_poses[ranked_poses.size()-1].mol, mol);

	  // find position with the highest score
	  max_score = ranked_poses[0].score;
	  max_position = 0;

	  for(i=0;i<ranked_poses.size();i++) {
		  if(ranked_poses[i].score > max_score)
			  max_position = i;
	  }

	  high_score = max_score;
	  high_position = max_position;

  } else if (mol.current_score < high_score) {

	  // add mol to the high position
	  ranked_poses[high_position].data = mol.current_data;
	  ranked_poses[high_position].score = mol.current_score;
	  copy_molecule(ranked_poses[high_position].mol, mol);

	  // find position with the highest score
	  max_score = ranked_poses[0].score;
	  max_position = 0;

	  for(i=0;i<ranked_poses.size();i++) {
		  if(ranked_poses[i].score > max_score)
			  max_position = i;
	  }

	  high_score = max_score;
	  high_position = max_position;
  }

}

/************************************************/
void Library_File::write_scored_poses() {
  int		i, j;
//int		insert_point;
  ofstream	ofs;
  SCOREMol	tmp_mol;
  float		crmsd;

  /**
  change if statement- only skip the writing if it is a client
  for both single proc. and master mode, writing should occur
  **/

	  if(ranked_poses.size() > 0) {

		// sort ranked_poses here
		sort(ranked_poses.begin(), ranked_poses.end());

		// cluster ranked poses if requested
		if(cluster_ranked_poses) {

			// initialize clustering data
			cluster_assignments.clear();
			cluster_assignments.resize(ranked_poses.size(), 1);

			// loop over ranked poses
			for(i=0;i<ranked_poses.size()-1;i++) {

				// if mol is not assigned to a cluster already
				if(cluster_assignments[i] > 0) {
					
					// loop over nested poses
					for(j=i+1;j<ranked_poses.size();j++) {

						if(cluster_assignments[j] > 0) {
							crmsd = calculate_rmsd(ranked_poses[i].mol, ranked_poses[j].mol);
					
							// if second mol is w/in threhsold of the first
							if(crmsd < cluster_rmsd_threshold) {
								cluster_assignments[i]++;
								cluster_assignments[j] = 0;
							}
						}
					}
				}
			}
		}
		// end clustering

		// print out the ligand info
		cout << endl << "-----------------------------------" << endl;
		cout << "Molecule: " << ranked_poses[0].mol.title << endl << endl;

		cout << " Anchors:\t\t" << num_anchors << endl;
		cout << " Orientations:\t\t" << num_orients << endl;
		cout << " Conformations:\t\t" << num_confs << endl;

		if(write_scored_ligands || rank_ligands) {
			cout << endl;
			if(ranked_poses[0].mol.num_atoms > 0)
				for(i=0;i<ranked_poses[0].data.size();i++)
					if(ranked_poses[0].data[i] != '#')
						cout << ranked_poses[0].data[i];
		}

		if(write_scored_ligands) {

			for(i=0;i<ranked_poses.size();i++) {

				if(ranked_poses[i].mol.num_atoms > 0) {

					// if no clustering is used- simply write out molecule
					if(!cluster_ranked_poses) {					

						ligand_out_scored << "\n########## Name:\t\t" << ranked_poses[i].mol.title << endl;

						if(calc_rmsd)
							ligand_out_scored << "########## RMSD:\t\t" << calculate_rmsd(ranked_poses[i].mol) << endl;

						ligand_out_scored << ranked_poses[i].data << endl;

						write_mol(ranked_poses[i].mol, ligand_out_scored);

					} else {

						// if clustering is used, and molecule is a clusterhead, write it out
						if(cluster_assignments[i] > 0) {
							ligand_out_scored << "\n########## Name:\t\t" << ranked_poses[i].mol.title << endl;

							if(calc_rmsd)
								ligand_out_scored << "########## RMSD:\t\t" << calculate_rmsd(ranked_poses[i].mol) << endl;
						
							ligand_out_scored << "########## Cluster Size:\t" << cluster_assignments[i] << endl;
							ligand_out_scored << ranked_poses[i].data << endl;
							write_mol(ranked_poses[i].mol, ligand_out_scored);
						}
					}

				}

			}
		}

		if(rank_ligands) {

			if(ranked_poses[0].mol.num_atoms > 0) {

				tmp_mol.first = ranked_poses[0].score;
				copy_molecule(tmp_mol.second, ranked_poses[0].mol);
				tmp_mol.second.score_text_data = ranked_poses[0].data;

				ranked_list.push_back(tmp_mol);

				if(ranked_list.size() > 2*max_ranked) {
					sort(ranked_list.begin(), ranked_list.end(), less_than_pair);

					while(ranked_list.size() > max_ranked)
						ranked_list.pop_back();
				}

			}
		}

		ranked_poses.clear();
		high_position = 0;
		high_score = 0.0;

	}


}

/************************************************/
void Library_File::write_ranked_ligands() {
	int	i;

	sort(ranked_list.begin(), ranked_list.end(), less_than_pair);

	while(ranked_list.size() > max_ranked)
		ranked_list.pop_back();

	if(rank_ligands) {
		for(i=0;i<ranked_list.size();i++) {
			ligand_out_ranked << "\n########## Name:\t\t" << ranked_list[i].second.title << endl;

			if(calc_rmsd)
				ligand_out_ranked << "########## RMSD:\t\t" << calculate_rmsd(ranked_list[i].second) << endl;

			ligand_out_ranked << ranked_list[i].second.score_text_data << endl;

			write_mol(ranked_list[i].second, ligand_out_ranked);

		}
	}

}

/************************************************/
void Library_File::submit_rmsd_reference(DOCKMol &mol) {
	ifstream	rmsd_ref;

		copy_molecule(rmsd_reference, mol);

}



/************************************************/
void Library_File::submit_rmsd_reference_file_open() {
//	ifstream	rmsd_ref;

		if(constant_rmsd_ref) {  
		rmsd_ref.open(constant_rmsd_ref_file.c_str());
//		Read_Mol2(rmsd_reference, rmsd_ref);
//		rmsd_ref.close();
		}
}

/************************************************/
void Library_File::submit_rmsd_reference_file_close() {
//	ifstream	rmsd_ref;

		if(constant_rmsd_ref) { 
//		rmsd_ref.open(constant_rmsd_ref_file.c_str());
//		Read_Mol2(rmsd_reference, rmsd_ref);
		rmsd_ref.close();
		}
}


/************************************************/
float Library_File::calculate_rmsd(DOCKMol &mol) {
	float	rmsd;
	int		i;
//	int		hcount;

	rmsd = 0.00;

	if(rmsd_reference.num_atoms != mol.num_atoms)
		return 99.99;

//	hcount = 0;

	
	rmsd = bstrot(rmsd_reference, mol);
	

//	for(i=0;i<mol.num_atoms;i++) {
//		if(mol.amber_at_heavy_flag[i]) {
//			rmsd +=( pow((mol.x[i] - rmsd_reference.x[i]),2) + pow((mol.y[i] - rmsd_reference.y[i]),2) + pow((mol.z[i] - rmsd_reference.z[i]),2) );
//			hcount++;
//		}
//	}

//	rmsd = rmsd/mol.num_atoms;
//	rmsd = rmsd/hcount;
//	rmsd = sqrt(rmsd);

	return rmsd;

}

/************************************************/
float Library_File::calculate_rmsd(DOCKMol &refmol, DOCKMol &mol) {
	float	rmsd;
	int		i;
	int		hcount;

	rmsd = 0.00;

	if(refmol.num_atoms != mol.num_atoms)
		return 99.99;

	hcount = 0;

	for(i=0;i<mol.num_atoms;i++) {
		if(mol.amber_at_heavy_flag[i]) {
			rmsd +=( pow((mol.x[i] - refmol.x[i]),2) + pow((mol.y[i] - refmol.y[i]),2) + pow((mol.z[i] - refmol.z[i]),2) );
			hcount++;			
		}
	}

//	rmsd = rmsd/mol.num_atoms;
	rmsd = rmsd/hcount;
	rmsd = sqrt(rmsd);

	return rmsd;
}

/************************************************/
float Library_File::calculate_Intra_Lig_Energy(DOCKMol &mol, Master_Conformer_Search &master) {
	float	total_energy;
	float	total_vdw;
	float	total_es; 
	float   vdwAi, vdwAj, vdwBi, vdwBj;
	double 	distance;
	
	int		i;
	int		j;

       ie_diel = master.c_ag_conf.ie_diel;
       ie_att_exp = master.c_ag_conf.ie_att_exp;
       ie_rep_exp = master.c_ag_conf.ie_rep_exp;
       
//	cout << "########## test_transfer in SubmConf ie_diel:\t\t" <<  ie_diel << endl;
	
	total_vdw = 0.00;
	total_es = 0.00;

	for(i=0;i<mol.num_atoms;i++) {
        	vdwAi = sqrt( mol.amber_at_well_depth[i] * (ie_att_exp/(ie_rep_exp-ie_att_exp)) * pow((2*mol.amber_at_radius[i]),ie_rep_exp) );
		vdwBi = sqrt( mol.amber_at_well_depth[i] * (ie_rep_exp/(ie_rep_exp-ie_att_exp)) * pow((2*mol.amber_at_radius[i]),ie_att_exp) );
	
//		cout << "########## charges i:\t\t" << mol.charges[i]  << endl;
	
		for(j=i+1;j<mol.num_atoms;j++) {	
		  vdwAj = sqrt( mol.amber_at_well_depth[j] * (ie_att_exp/(ie_rep_exp-ie_att_exp)) * pow((2*mol.amber_at_radius[j]),ie_rep_exp) );
		  vdwBj = sqrt( mol.amber_at_well_depth[j] * (ie_rep_exp/(ie_rep_exp-ie_att_exp)) * pow((2*mol.amber_at_radius[j]),ie_att_exp) );
//			if(mol.amber_at_heavy_flag[i]) {
//				if(mol.amber_at_heavy_flag[j]) {
				   	if((mol.get_bond(i,j) == -1) && (!mol.atoms_are_one_three(i,j)) && (!mol.atoms_are_one_four(i,j))) {

						distance = pow((mol.x[i]-mol.x[j]),2) + pow((mol.y[i]-mol.y[j]),2) + pow((mol.z[i]-mol.z[j]),2);
						if(distance < 1.0)
							distance = 1.0;
						distance = sqrt(distance);

						total_vdw += (vdwAi*vdwAj) / pow(distance, (ie_rep_exp)) - (vdwBi*vdwBj) / pow(distance, (ie_att_exp));
						total_es  += (mol.charges[i] * mol.charges[j]) / (ie_diel * distance);	
					}			

//				}

	
//			}
		
		}
	}
	
//	cout << "########## Ligand_Conf_En vdw:\t\t" << total_vdw << endl;	
//	cout << "########## Ligand_Conf_En es:\t\t" << total_es << endl;	
	
	total_energy = total_vdw + total_es;
//	cout << "########## Ligand_Conf_En:\t\t" << total_energy << endl;	
	return total_energy;
}




/************************************************/
/* routine to superimpose two sets of coordinates
*  
* cadged from Larry Andrews' Newvector.f library
*
C  The algorithm is from Ferro and Hermans, Acta Cryst., A33,345-347
C  (1977).  Other algorithms give the same result, such as:
C
C  Kearsley, Acta Cryst., A45, 208-210 (1989)
C  Diamond,  Acta Cryst., A44, 211-216 (1988)
C  Kabsch,   Acta Cryst., A34, 827-828 (1978)
C  Kabsch,   Acta Cryst., A32, 922-923 (1976)
C
C  The algorithm of Ferro and Hermans has the advantage that the
C  rotational components are removed iteratively, so that residual
C  numerical errors are correctly removed.  This leaves no
C  possibility of poorly orthogonalized matrices.  The explicit
C  iteration is not a problem.  The other methods simply hide the
C  iteration in the eigenvector solver.  Ferro and Hermans
C  simply build what is a simple, special purpose version of the QR
C  method.
*/

//include <math.h>
/*void matmul(float[],float[],float[],int,int);
*/

/* return the RMS error, -1 on internal error */
float Library_File::bstrot(DOCKMol &refmol, DOCKMol &mol) {
  int na;
  float matrix[3][3], delta[3];

    float tensor[3][3],cx,cy,cz,cxx,cyy,czz;
    float tx,ty,tz,txx,tyy,tzz;
    float rms;
    float sx[3][3],sy[3][3],sz[3][3];
    float sq[3][3];

    int i,j,ipass, hcount;
 
 	na = mol.num_atoms;
	if(refmol.num_atoms != mol.num_atoms)
		return 99.99;


    if( na < 1) return -1.;
    /* find the centers of mass */
    cx = 0.;
    cy = 0.;
    cz = 0.;
    cxx = 0.;
    cyy = 0.;
    czz = 0.;
    for( i=0; i< na; i++)
    {
        cx += mol.x[i];
        cy += mol.y[i];
        cz += mol.z[i];
        cxx += refmol.x[i];
        cyy += refmol.y[i];
        czz += refmol.z[i];
    }
    cx /= na;
    cy /= na;
    cz /= na;
    cxx /= na;
    cyy /= na;
    czz /= na;
    /* make the metric tensor */
    for( i=0; i< 3; i++)
        for( j=0; j< 3; j++)
        {
            tensor[i][j] = 0.;
            matrix[i][j] = 0.;
            sx[i][j] = 0.;
            sy[i][j] = 0.;
            sz[i][j] = 0.;
        }
    matrix[0][0] = 1.;
    matrix[1][1] = 1.;
    matrix[2][2] = 1.;
    sx[0][0] = 1.;
    sy[1][1] = 1.;
    sz[2][2] = 1.;
    for( i=0; i<na; i++)
    {
        tx = mol.x[i] - cx;
        ty = mol.y[i] - cy;
        tz = mol.z[i] - cz;
        txx = refmol.x[i] - cxx;
        tyy = refmol.y[i] - cyy;
        tzz = refmol.z[i] - czz;
        tensor[0][0] += tx*txx;
        tensor[0][1] += tx*tyy;
        tensor[0][2] += tx*tzz;
        tensor[1][0] += ty*txx;
        tensor[1][1] += ty*tyy;
        tensor[1][2] += ty*tzz;
        tensor[2][0] += tz*txx;
        tensor[2][1] += tz*tyy;
        tensor[2][2] += tz*tzz;
    }
    /* now find the linear orthogonal transformation which symetrizes
    * the metric tensor */

    for( ipass = 0; ipass < 20; ipass ++)
    {
        rms = 0.;
        /* x */
        tx = atan2( tensor[2][1]-tensor[1][2],
                    tensor[1][1]+tensor[2][2]);
        rms += fabs(tx);
        ty = cos(tx); tz = sin(tx);
        sx[1][1] = ty;
        sx[2][1] = -tz;
        sx[1][2] = tz;
        sx[2][2] = ty;
        matmul(sx, tensor, sq);
        cpyvec(sq, tensor);
        matmul( sx,matrix,sq);
        cpyvec(sq, matrix);
        /* y */
        tx = atan2( tensor[2][0]-tensor[0][2],
                    tensor[0][0]+tensor[2][2]);
        rms += fabs(tx);
        ty = cos(tx); tz = sin(tx);
        sy[0][0] = ty;
        sy[2][0] = -tz;
        sy[0][2] = tz;
        sy[2][2] = ty;
        matmul( sy,tensor,sq);
        cpyvec(sq, tensor);
        matmul( sy,matrix,sq);
        cpyvec(sq, matrix);
        /* z */
        tx = atan2( tensor[0][1]-tensor[1][0],
                    tensor[1][1]+tensor[0][0]);
        rms += fabs(tx);
        ty = cos(tx); tz = sin(tx);
        sz[1][1] = ty;
        sz[0][1] = -tz;
        sz[1][0] = tz;
        sz[0][0] = ty;
        matmul( sz,tensor,sq);
        cpyvec(sq, tensor);
        matmul( sz,matrix,sq);
        cpyvec(sq, matrix);
        /*  termination critereon here */
        if( rms < 1.e-7) break;
    }
    rms = 0.;
    hcount = 0.;
    for(i=0; i< na; i++)
    {
        txx = refmol.x[i] - cxx;
        tyy = refmol.y[i] - cyy;
        tzz = refmol.z[i] - czz;
        tx = matrix[0][0]*txx + matrix[1][0]*tyy + matrix[2][0]*tzz;
        ty = matrix[0][1]*txx + matrix[1][1]*tyy + matrix[2][1]*tzz;
        tz = matrix[0][2]*txx + matrix[1][2]*tyy + matrix[2][2]*tzz;
        tx += cx - mol.x[i];
        ty += cy - mol.y[i];
        tz += cz - mol.z[i];
	if(mol.amber_at_heavy_flag[i])	{
	        rms += tx*tx + ty*ty + tz*tz;
		hcount++;
	}
    }
    tx = matrix[0][0]*cxx + matrix[1][0]*cyy + matrix[2][0]*czz;
    ty = matrix[0][1]*cxx + matrix[1][1]*cyy + matrix[2][1]*czz;
    tz = matrix[0][2]*cxx + matrix[1][2]*cyy + matrix[2][2]*czz;
    delta[0] = cx - tx;
    delta[1] = cy - ty;
    delta[2] = cz - tz;
    if(hcount == 0 ) 
    	return 99.99;
    return sqrt(rms/hcount);


}/* end of routine */
/* copy a vector into another */
void Library_File::cpyvec(float orig[3][3], float copy[3][3]) {
    int i,j;
         
    for( i=0; i< 3; i++) {
    	for( j=0; j< 3; j++) {
        copy[i][j] = orig[i][j];
	}
    }
	
}	
/*
multiply c[n][n] =  a[n][m] b[m][n];
*/
void Library_File::matmul( float a[3][3],float b[3][3],float c[3][3]) {

  int n,m;

    int i,j,k,ioff,koff;
	n = 3;
	m = 3;	
	
    for(i=0;i< n;i++)
	for(j=0;j< n;j++)
	        c[i][j] = 0.;
    for(i=0;i< n;i++) {
//        ioff = i*n;
        for(j=0;j<n;j++) {
//            koff = 0.;
            for(k=0;k<m;k++) {
//                c[ ioff +j] += a[ioff + k] *b[ j +koff];
                c[i][j] += a[i][k] *b[k][j];
//                koff += m;
            }
        }
    }

}


// ------end superimpose --------------

