using namespace std;

/**************************************************/
class Library_File : public Base_MPI {

 public:
  // data to identify database type (mol2 or heirarchy)
  int				db_type;		// 1=MOL2 (*.mol2), 2=HDB (*.db)
  vector<DOCKMol>	mol_branches;	// vector for the branches of an HDB molecule
  vector<DOCKMol> 	mol_conf_printed;
  DOCKMol               initial_mol;
  DOCKMol		mol_ref_ext;

  // general lib class data members
  ifstream			ligand_in;
  ifstream			rmsd_ref;

  ofstream			ligand_out_orients;
  ofstream			ligand_out_scored;
  ofstream			ligand_out_confs;
  ofstream			ligand_out_ranked;

  bool				max_mol_limit;
  int				max_mols;
  int				initial_skip;
  int				total_mols;
  string			input_file_name;
  string			output_file_prefix;
  string			output_file_orient;
  string			output_file_confs;
  string			output_file_scored;
  string			output_file_ranked;

  DOCKMol			rmsd_reference;
  bool				calc_rmsd;
  int				num_anchors;
  int				num_orients;
  int				num_confs;
  bool				constant_rmsd_ref;
  string			constant_rmsd_ref_file;

  DOCKMol			bestmol;
  float				bestscore;
  string			bestdata;

  vector<RANKMol>	ranked_poses;
  int				high_position;
  float				high_score;
  int				num_scored_poses;

  bool				write_orients;
  bool				write_conformers;
  bool				write_scored_ligands;

  bool				read_success;

  bool				rank_ligands;
  vector<SCOREMol>	ranked_list;
  int				max_ranked;
  bool				rank_ligand_output_override;
  float				rank_list_highest_energy;
  int				rank_list_highest_pos;
  int				completed;
  
  // final pose RMSD clustering
  bool				cluster_ranked_poses;
  INTVec			cluster_assignments;
  float				cluster_rmsd_threshold;

  void				input_parameters(Parameter_Reader &parm, bool, int &);
  void				initialize(int, char **, bool);
  void				open_files();
  void				close_files();

  bool				read_mol(DOCKMol &);
  void				write_mol(DOCKMol &, ofstream &);
  bool				get_mol(DOCKMol &, bool);

  void				submit_orientation(DOCKMol &, Master_Score &);
  void				submit_conformation(DOCKMol &, DOCKMol &, vector<DOCKMol> &, Master_Score &, Master_Conformer_Search &, int &);
  void				submit_scored_pose(DOCKMol &, Master_Score &);
  void				write_scored_poses();
  void				write_ranked_ligands();
  void				old_submit_scored_pose(DOCKMol &, Master_Score &);
  void				old_write_scored_poses();

  void				submit_rmsd_reference(DOCKMol &);
  void				submit_rmsd_reference_file_open();
  void				submit_rmsd_reference_file_close();
  float				calculate_rmsd(DOCKMol &);
  float				calculate_rmsd(DOCKMol &, DOCKMol &);
  float 			segment_internal_energy(DOCKMol &mol, int layer_num, int current_bond);
  float				calculate_Intra_Lig_Energy(DOCKMol &, Master_Conformer_Search &);
  float				bstrot(DOCKMol &, DOCKMol &);
  void				cpyvec(float orig[3][3], float copy[3][3]);
  void				matmul(float a[3][3], float b[3][3], float c[3][3]);
  float				*vdwA;
  float				*vdwB;
  int				ie_att_exp;
  int				ie_rep_exp;
  float				ie_diel;
  bool				use_internal_energy;
  float				energy_window;
  float				rmsd_window;
  float 			ref_en;
  int				printed_conf;
  int				max_conf;
  string			original_mol_title;

  // heirarchy DB functions
  string			hdb_atom_type_converter(int);
  bool				read_hierarchy_db(DOCKMol &, ifstream &);

};

