#include "dock.h"

/******************************************************/
void Master_Score::input_parameters(Parameter_Reader &parm) {
	string	tmp;
//	bool	use_score;

//	cout << "\nMaster Score Parameters" << endl; com by mmiteva
//	cout << "------------------------------------------------------------------------------------------" << endl;

//	tmp = parm.query_param("score_molecules","yes","yes no");
	tmp = "no";
	if(tmp == "yes")
		use_score = true;
	else
		use_score = false;

	if(use_score) {
		primary_score_found = false;
		secondary_score_found = false;
		use_primary_score = false;
		use_secondary_score = false;

		c_cnt.use_score = false;
		c_cnt.use_primary_score = false;
		c_cnt.use_secondary_score = false;
		c_nrg.use_score = false;
		c_nrg.use_primary_score = false;
		c_nrg.use_secondary_score = false;
		c_cont_nrg.use_score = false;
		c_cont_nrg.use_primary_score = false;
		c_cont_nrg.use_secondary_score = false;
		c_gbsa.use_score = false;
		c_gbsa.use_primary_score = false;
		c_gbsa.use_secondary_score = false;
		c_gbsa_hawkins.use_primary_score = false;
		c_gbsa_hawkins.use_secondary_score = false;

		if((!primary_score_found)||(!secondary_score_found)) {
			c_nrg.input_parameters(parm, primary_score_found, secondary_score_found);
		}

		if((!primary_score_found)||(!secondary_score_found)) {
			c_cont_nrg.input_parameters(parm, primary_score_found, secondary_score_found);
		}

		if((!primary_score_found)||(!secondary_score_found)) {
			c_cnt.input_parameters(parm, primary_score_found, secondary_score_found);
		}

//		if((!primary_score_found)||(!secondary_score_found)) {
//			c_gbsa_hawkins.input_parameters(parm, primary_score_found, secondary_score_found);
//		}

		if((!primary_score_found)||(!secondary_score_found)) {
			c_gbsa.input_parameters(parm, primary_score_found, secondary_score_found);
		}

		if(!primary_score_found) {
			cout << "\nError:  No scoring function selected for the primary score.  Program will now terminate.\n" << endl;
			exit(0);
		} else
			use_primary_score = true;

		if(!secondary_score_found) {
			cout << "\nWarning:  No secondary scoring function selected.  Primary scoring function will be used as secondary.\n" << endl;
			use_secondary_score = false;
		} else
			use_secondary_score = true;
	} else {
		use_primary_score = false;
		use_secondary_score = false;
	}

}

/******************************************************/
void Master_Score::initialize_all(AMBER_TYPER &typer, int argc, char **argv) {

	if(use_primary_score) {

		if(c_cnt.use_score == true) {
			c_cnt.initialize();

			if(c_cnt.use_primary_score)
				primary_score = &c_cnt;
			if(c_cnt.use_secondary_score)
				secondary_score = &c_cnt;
		}

		if(c_nrg.use_score == true) {
			c_nrg.initialize(typer);

			if(c_nrg.use_primary_score)
				primary_score = &c_nrg;
			if(c_nrg.use_secondary_score)
				secondary_score = &c_nrg;
		}

		if(c_cont_nrg.use_score == true) {
			c_cont_nrg.initialize(typer);

			if(c_cont_nrg.use_primary_score)
				primary_score = &c_cont_nrg;
			if(c_cont_nrg.use_secondary_score)
				secondary_score = &c_cont_nrg;
		}

		if(c_gbsa.use_score == true) {
			c_gbsa.initialize(typer, argc, argv);

			if(c_gbsa.use_primary_score)
				primary_score = &c_gbsa;
			if(c_gbsa.use_secondary_score)
				secondary_score = &c_gbsa;
		}
/*
		if(c_gbsa_hawkins.use_score == true) {
			c_gbsa_hawkins.initialize(typer);

			if(c_gbsa_hawkins.use_primary_score)
				primary_score = &c_gbsa_hawkins;
			if(c_gbsa_hawkins.use_secondary_score)
				secondary_score = &c_gbsa_hawkins;
		}
*/
		if(!use_secondary_score) {
			secondary_score = primary_score;
			use_secondary_score = true;
		}

	}

}

/******************************************************/
bool Master_Score::compute_primary_score(DOCKMol &mol) {

	if(use_primary_score)
		return primary_score->compute_score(mol);
	else
		return true;

}

/******************************************************
float Master_Score::return_primary_score() {

	if(use_primary_score)
		return primary_score->return_current_score();
	else
		return 0.0;

}
*/

/******************************************************/
bool Master_Score::compute_secondary_score(DOCKMol &mol) {

	if(use_secondary_score)
		return secondary_score->compute_score(mol);
	else
		return true;

}

/******************************************************
float Master_Score::return_secondary_score() {

	if(use_secondary_score)
		return secondary_score->return_current_score();
	else
		return 0.0;

}
*/

