#include "dock.h"

/************************************************/
void Bump_Filter::input_parameters(Parameter_Reader &parm) {
  string tmp;

//  cout << "\nBump Filter Parameters" << endl; com by mmiteva
//  cout << "------------------------------------------------------------------------------------------" << endl;

//  tmp = parm.query_param("bump_filter","yes","yes no"); com by mmiteva
  tmp = "no";

  if(tmp == "yes")
	  bump_filter = 1;
  else
	  bump_filter = 0;

  if(bump_filter == 1) {
	  file_prefix = parm.query_param("bump_grid_prefix","grid");
	  bump_max    = atoi(parm.query_param("max_bumps", "2").c_str());
  }

}

/************************************************/
void Bump_Filter::initialize() {

  if(bump_filter) {
	  cout << "Initializing Bump Filter Routines..." << endl;
	  read_bump_grid();
	  calc_corner_coords();
	  use_score = true;
  } else
	  use_score = false;

}

/************************************************/
void Bump_Filter::read_bump_grid() {
  string	fname;

  fname = file_prefix + ".bmp";
  grid_in = fopen(fname.c_str(),"rb");

  if(grid_in == NULL) {
	  cout << "\n\nCould not open " << fname << " for reading.  Program will terminate." << endl << endl;
	  exit(0);
  }

  cout << " Reading the bump grid from " << fname << endl;

  fread (&size, sizeof (int), 1, grid_in);
  fread (&spacing, sizeof (float), 1, grid_in);
  fread (origin, sizeof (float), 3, grid_in);
  fread (span, sizeof (int), 3, grid_in);

  bump = (unsigned char *)malloc(size * sizeof(unsigned char));
  fread (bump, sizeof(unsigned char), size, grid_in);

  fclose(grid_in);

}

/************************************************/
bool Bump_Filter::check_bumps(DOCKMol &mol, bool bypass) {
  int num_bumps;

  if(bypass == false)
	  return true;

  if(bump_filter == 1) {

    num_bumps = get_bump_score(mol);

    if((num_bumps <= bump_max)&&(num_bumps != -1))
	  return true;
    else
	  return false;
  } else
	  return true;

}

/************************************************/
int Bump_Filter::get_bump_score(DOCKMol &mol) {
  int		bump_count, vdw_threshold;
  int		atom;

  bump_count = 0;
  vdw_threshold = 0;

  for(atom=0;atom<mol.num_atoms;atom++) {
	  if(mol.atom_active_flags[atom]) {
		  if(check_box_boundaries(mol.x[atom], mol.y[atom], mol.z[atom])) {
			  find_grid_neighbors(mol.x[atom], mol.y[atom], mol.z[atom]);
			  vdw_threshold = NINT( 10 * mol.amber_at_radius[atom] );

			  if( vdw_threshold >= (int)bump[nearest_neighbor] )
				  bump_count++;
		  } else
			  return -1;
	  }
  }

  return (bump_count);
}


/************************************************/
/************************************************/
/************************************************/
void Energy_Score::input_parameters(Parameter_Reader &parm, bool &primary_score, bool &secondary_score) {
  string tmp;

  use_primary_score = false;
  use_secondary_score = false;

  cout << "\nEnergy Score Parameters" << endl;
  cout << "------------------------------------------------------------------------------------------" << endl;

  if(!primary_score) {
	  tmp = parm.query_param("grid_score_primary","yes","yes no");
	  if(tmp == "yes")
		  use_primary_score = true;
	  else
		  use_primary_score = false;

	  primary_score = use_primary_score;
  }

  if(!secondary_score) {
	  tmp = parm.query_param("grid_score_secondary","yes","yes no");
	  if(tmp == "yes")
		  use_secondary_score = true;
	  else
		  use_secondary_score = false;

	  secondary_score = use_secondary_score;
  }

  if(use_primary_score || use_secondary_score)
	  energy_score = 1;
  else
	  energy_score = 0;

  if(energy_score == 1) {
	  vdw_scale  = atof(parm.query_param("grid_score_vdw_scale","1").c_str());
	  es_scale   = atof(parm.query_param("grid_score_es_scale","1").c_str());
	  file_prefix = parm.query_param("grid_score_grid_prefix","grid");
	  use_score = true;
  }
}

/************************************************/
void Energy_Score::initialize(AMBER_TYPER &typer) {

  if(energy_score == 1) {
	  vdwA_allocated = false;
	  vdwB_allocated = false;

	  cout << "Initializing Energy Score Routines..." << endl;

	  read_nrg_grid();
	  init_vdw_energy(typer);
	  calc_corner_coords();
	  use_score = true;
  } else
	  use_score = false;


}

/************************************************/
void Energy_Score::read_nrg_grid() {
  string	fname;

  fname = file_prefix + ".nrg";
  grid_in = fopen(fname.c_str(),"rb");

  if(grid_in == NULL) {
	  cout << "\n\nCould not open " << fname << " for reading.  Program will terminate." << endl << endl;
	  exit(0);
  }

  cout << " Reading the energy grid from " << fname << endl;

  fread (&size,       sizeof(int), 1, grid_in);
  fread (&atom_model, sizeof(int), 1, grid_in);
  fread (&att_exp,    sizeof(int), 1, grid_in);
  fread (&rep_exp,    sizeof(int), 1, grid_in);

  avdw = (float *)malloc(size * sizeof(float));
  bvdw = (float *)malloc(size * sizeof(float));
  es   = (float *)malloc(size * sizeof(float));

  fread ((void *)bvdw, sizeof(float), size, grid_in);
  fread ((void *)avdw, sizeof(float), size, grid_in);
  fread ((void *)es,   sizeof(float), size, grid_in);

  fclose(grid_in);

  fname = file_prefix + ".bmp";
  grid_in = fopen(fname.c_str(),"rb");

  if(grid_in == NULL) {
	  cout << "\n\nCould not open " << fname << " for reading.  Program will terminate." << endl << endl;
	  exit(0);
  }

  fread (&size, sizeof (int), 1, grid_in);
  fread (&spacing, sizeof (float), 1, grid_in);
  fread (origin, sizeof (float), 3, grid_in);
  fread (span, sizeof (int), 3, grid_in);

  fclose(grid_in);
}

/************************************************/
void Energy_Score::init_vdw_energy(AMBER_TYPER &typer) {
  int		i;

  if(vdwA_allocated) {
	  free(vdwA);
	  vdwA_allocated = false;
  }

  if(vdwB_allocated) {
	  free(vdwB);
	  vdwB_allocated = false;
  }

  vdwA = (float *)malloc(typer.d4_atom_typer.types.size()*sizeof(float));
  vdwA_allocated = true;
  vdwB = (float *)malloc(typer.d4_atom_typer.types.size()*sizeof(float));
  vdwB_allocated = true;

  for(i=0;i<typer.d4_atom_typer.types.size();i++) {
	  vdwA[i] = sqrt( typer.d4_atom_typer.types[i].well_depth * (att_exp/(rep_exp-att_exp)) * pow((2*typer.d4_atom_typer.types[i].radius), rep_exp) );
	  vdwB[i] = sqrt( typer.d4_atom_typer.types[i].well_depth * (rep_exp/(rep_exp-att_exp)) * pow((2*typer.d4_atom_typer.types[i].radius), att_exp) );
  }

}

/************************************************/
bool Energy_Score::compute_score(DOCKMol &mol) {
  float		vdw_val, es_val, total;
  int		atom;
  string	text;
  char		line[1000];


  total = vdw_val = es_val = 0.0;

  if(energy_score == 1) {

	  //check to see if molecule is inside grid box
	  for(atom=0;atom<mol.num_atoms;atom++){
			if(!check_box_boundaries(mol.x[atom], mol.y[atom], mol.z[atom])){
				mol.current_score = 1000000.0;
				sprintf(line, "ERROR:  Conformation could not be scored by DOCK.\nConformation not completely within grid box.\n");
				mol.current_data = line;
				return false;
			}
	  }

	  for(atom=0;atom<mol.num_atoms;atom++) {

		  if(mol.atom_active_flags[atom]) {

				  find_grid_neighbors(mol.x[atom], mol.y[atom], mol.z[atom]);

				  vdw_val += ((vdwA[mol.amber_at_id[atom]] * interpolate(avdw)) - (vdwB[mol.amber_at_id[atom]] * interpolate(bvdw)))*vdw_scale;
				  es_val  += mol.charges[atom] * interpolate(es) * es_scale;
			  } 


	  }

	  total = vdw_val + es_val;

	  vdw_component = vdw_val;
	  es_component = es_val;

	  mol.current_score = total;

//	  add_score_to_mol(mol, "Energy Score", total);

  }

  return true;
}

/************************************************/
string Energy_Score::output_score_summary(float score) {
  string	text;
  char		line[1000];

  if(use_score) {

	  sprintf(line, "########## Energy Score:%20f\n##########%14s%20f\n##########%14s%20f\n##########\n", score,"vdw:", vdw_component, "es:", es_component);
	  text = line;

	  return text;
  } else
	  return "";
}

/************************************************/
/************************************************/
/************************************************/
void Continuous_Energy_Score::input_parameters(Parameter_Reader &parm, bool &primary_score, bool &secondary_score) {
  string tmp;

  use_primary_score = false;
  use_secondary_score = false;

  cout << "\nContinuous Energy Score Parameters" << endl;
  cout << "------------------------------------------------------------------------------------------" << endl;

  if(!primary_score) {
	  tmp = parm.query_param("continuous_score_primary","no","yes no");
	  if(tmp == "yes")
		  use_primary_score = true;
	  else
		  use_primary_score = false;

	  primary_score = use_primary_score;
  }

  if(!secondary_score) {
	  tmp = parm.query_param("continuous_score_secondary","no","yes no");
	  if(tmp == "yes")
		  use_secondary_score = true;
	  else
		  use_secondary_score = false;

	  secondary_score = use_secondary_score;
  }

  if(use_primary_score || use_secondary_score)
	  use_score = true;
  else
	  use_score = false;

  if(use_score) {
	  receptor_filename = parm.query_param("cont_score_rec_filename", "receptor.mol2");

	  att_exp = atoi(parm.query_param("cont_score_att_exp", "6").c_str());
	  rep_exp = atoi(parm.query_param("cont_score_rep_exp", "12").c_str());
	  diel_screen = atof(parm.query_param("cont_score_dielectric", "4.0").c_str());

	  vdw_scale  = atoi(parm.query_param("cont_score_vdw_scale","1").c_str());
	  es_scale   = atoi(parm.query_param("cont_score_es_scale","1").c_str());

  }
}

/************************************************/
void Continuous_Energy_Score::initialize(AMBER_TYPER &typer) {
  ifstream		rec_file;

  if(use_score) {

	  vdwA_allocated = false;
	  vdwB_allocated = false;
	  init_vdw_energy(typer);

	  rec_file.open(receptor_filename.c_str());

	  if(rec_file.fail()) {
		  cout << "Error Opening Receptor File!" << endl;
		  exit(0);
	  }

	  if(!Read_Mol2(receptor, rec_file)) {
		  cout << "Error Reading Receptor Molecule!" << endl;
		  exit(0);
	  }

	  rec_file.close();

	  typer.prepare_molecule(receptor);

  }
}


/************************************************/
void Continuous_Energy_Score::init_vdw_energy(AMBER_TYPER &typer) {
  int		i;

  if(vdwA_allocated) {
	  free(vdwA);
	  vdwA_allocated = false;
  }

  if(vdwB_allocated) {
	  free(vdwB);
	  vdwB_allocated = false;
  }

  vdwA = (float *)malloc(typer.d4_atom_typer.types.size()*sizeof(float));
  vdwA_allocated = true;
  vdwB = (float *)malloc(typer.d4_atom_typer.types.size()*sizeof(float));
  vdwB_allocated = true;

  //(att_exp/(rep_exp-att_exp))
  //(rep_exp/(rep_exp-att_exp))

  for(i=0;i<typer.d4_atom_typer.types.size();i++) {
	  vdwA[i] = sqrt( typer.d4_atom_typer.types[i].well_depth * pow((2*typer.d4_atom_typer.types[i].radius), rep_exp) );
	  vdwB[i] = sqrt( 2*typer.d4_atom_typer.types[i].well_depth * pow((2*typer.d4_atom_typer.types[i].radius), att_exp) );
  }

}

/************************************************/
bool Continuous_Energy_Score::compute_score(DOCKMol &mol) {
  float		vdw_val, es_val, total;
  int		i,j;
  float		dist;

  total = vdw_val = es_val = 0.0;

  if(use_score == 1) {

	  for(i=0;i<mol.num_atoms;i++) {

		  for(j=0;j<receptor.num_atoms;j++) {

			  if(mol.atom_active_flags[i]) {

			    dist = sqrt(pow((mol.x[i]-receptor.x[j]),2) + pow((mol.y[i]-receptor.y[j]),2) + pow((mol.z[i]-receptor.z[j]),2));
				vdw_val += (((vdwA[mol.amber_at_id[i]] * vdwA[receptor.amber_at_id[j]])/pow(dist, rep_exp)) - ((vdwB[mol.amber_at_id[i]] * vdwB[receptor.amber_at_id[j]])/pow(dist, att_exp)))*vdw_scale;
				es_val  += ((332 * mol.charges[i] * receptor.charges[j])/(pow(dist,2) * diel_screen)) * es_scale;

//				cout << vdwA[mol.amber_at_id[i]] << "\t" << vdwA[receptor.amber_at_id[j]] << "\t";
//				cout << vdwB[mol.amber_at_id[i]] << "\t" << vdwB[receptor.amber_at_id[j]] << "\t\t";
//				cout << vdw_val << "\t" << es_val << endl;
			  }

		  }

	  }

	  total = vdw_val + es_val;

	  vdw_component = vdw_val;
	  es_component = es_val;

	  mol.current_score = total;

//	  add_score_to_mol(mol, "Energy Score", total);

  }

  return true;
}

/************************************************/
string Continuous_Energy_Score::output_score_summary(float score) {
  string	text;
  char		line[1000];

  if(use_score) {
	  sprintf(line, "########## Energy Score:%20f\n##########%14s%20f\n##########%14s%20f\n##########\n", score,"vdw:", vdw_component, "es:", es_component);
	  text = line;

	  return text;
  } else
	  return "";
}

/************************************************/
/************************************************/
/************************************************/
void Contact_Score::input_parameters(Parameter_Reader &parm, bool &primary_score, bool &secondary_score) {
  string	tmp;

  use_primary_score = false;
  use_secondary_score = false;

  cout << "\nContact Score Paramters" << endl;
  cout << "------------------------------------------------------------------------------------------" << endl;

  if(!primary_score) {
	  tmp = parm.query_param("contact_score_primary","no","yes no");
	  if(tmp == "yes")
		  use_primary_score = true;
	  else
		  use_primary_score = false;

	  primary_score = use_primary_score;
  }

  if(!secondary_score) {
	  tmp = parm.query_param("contact_score_secondary","no","yes no");
	  if(tmp == "yes")
		  use_secondary_score = true;
	  else
		  use_secondary_score = false;

	  secondary_score = use_secondary_score;
  }

  if(use_primary_score || use_secondary_score)
	  contact_score = 1;
  else
	  contact_score = 0;

  if(contact_score == 1) {
	  use_score = true;
	  cutoff_distance = atof(parm.query_param("contact_score_cutoff_distance","4.5").c_str());
	  clash_overlap   = atof(parm.query_param("contact_score_clash_overlap","0.75").c_str());
	  clash_penalty   = atof(parm.query_param("contact_score_clash_penalty","50").c_str());
	  file_prefix = parm.query_param("contact_score_grid_prefix","grid");
  }

}

/************************************************/
void Contact_Score::initialize() {

	if(contact_score == 1) {
		cout << "Initializing Contact Score Routines..." << endl;
        read_cnt_grid();
		calc_corner_coords();
		use_score = true;
	} else
		use_score = false;

}

/************************************************/
void Contact_Score::read_cnt_grid() {
  string	fname;

  fname = file_prefix + ".cnt";
  grid_in = fopen(fname.c_str(),"rb");

  if(grid_in == NULL) {
	  cout << "\n\nCould not open " << fname << " for reading.  Program will terminate." << endl << endl;
	  exit(0);
  }

  cout << " Reading the contact grid from " << fname << endl;

  fread (&size, sizeof(int), 1, grid_in);
  cnt = (short int *)malloc(size * sizeof(short int));
  fread (cnt, sizeof(short int), size, grid_in);

  fclose(grid_in);

  fname = file_prefix + ".bmp";
  grid_in = fopen(fname.c_str(),"rb");

  if(grid_in == NULL) {
	  cout << "\n\nCould not open " << fname << " for reading.  Program will terminate." << endl << endl;
	  exit(0);
  }

  fread (&size, sizeof (int), 1, grid_in);
  fread (&spacing, sizeof (float), 1, grid_in);
  fread (origin, sizeof (float), 3, grid_in);
  fread (span, sizeof (int), 3, grid_in);

  bump = (unsigned char *)malloc(size * sizeof(unsigned char));
  fread (bump, sizeof(unsigned char), size, grid_in);

  fclose(grid_in);

}

/************************************************/
bool Contact_Score::compute_score(DOCKMol &mol) {
  float		cnt_score;
  int		vdw_threshold;
  int		atom;
  char		line[1000];

  cnt_score = 0;
  vdw_threshold = 0;

  if(contact_score == 1) {	  
	  
	  for(atom=0;atom<mol.num_atoms;atom++){
			if(!check_box_boundaries(mol.x[atom], mol.y[atom], mol.z[atom])){
				mol.current_score = 0.0;
				sprintf(line, "ERROR:  Conformation could not be scored by DOCK.\nConformation not completely within grid box.\n");
				mol.current_data = line;
				return false;
			}
	  }

	  for(atom=0;atom<mol.num_atoms;atom++) {

		  if(mol.amber_at_heavy_flag[atom]) {

			  if(mol.atom_active_flags[atom]) {

					  find_grid_neighbors(mol.x[atom], mol.y[atom], mol.z[atom]);

					  vdw_threshold = NINT( 10 * mol.amber_at_radius[atom] );

					  if(vdw_threshold >= (int)bump[nearest_neighbor]) {
						  cnt_score += clash_penalty;
					  } else {
						  cnt_score += (float)cnt[nearest_neighbor];
					  }

			  }

		  }

	  }

	  mol.current_score = cnt_score;

  }

  return true;
}

/************************************************/
string Contact_Score::output_score_summary(float score) {
  string	text;
  char		line[1000];

  if(use_score) {
	  sprintf(line, "########## Contact Score:%19f\n##########\n",score);
	  text = line;

	  return text;
  } else
	  return "";
}

/************************************************/
/************************************************/
/************************************************/
float square_distance (float a[3], float b[3])
{
	return
		(a[0] - b[0]) * (a[0] - b[0]) +
		(a[1] - b[1]) * (a[1] - b[1]) +
		(a[2] - b[2]) * (a[2] - b[2]);
}

/************************************************/
void GB_Pairwise::input_parameters(Parameter_Reader &parm, bool &primary_score, bool &secondary_score) {
  string	tmp;

  use_primary_score = false;
  use_secondary_score = false;

  cout << "\nGB/SA Score Parameters" << endl;
  cout << "------------------------------------------------------------------------------------------" << endl;

  if(!primary_score) {
	  tmp = parm.query_param("gbsa_pairwise_score_primary","no","yes no");
	  if(tmp == "yes")
		  use_primary_score = true;
	  else
		  use_primary_score = false;

	  primary_score = use_primary_score;
  }

  if(!secondary_score) {
	  tmp = parm.query_param("gbsa_pairwise_score_secondary","no","yes no");
	  if(tmp == "yes")
		  use_secondary_score = true;
	  else
		  use_secondary_score = false;

	  secondary_score = use_secondary_score;
  }

  if(use_primary_score || use_secondary_score)
	  use_score = 1;
  else
	  use_score = 0;

  if(use_score == 1) {
	  gb_grid_prefix = parm.query_param("gbsa_pairwise_gb_grid_prefix","gb_grid");
	  sa_grid_prefix = parm.query_param("gbsa_pairwise_sa_grid_prefix","sa_grid");
          vdw_grid_prefix = parm.query_param("gbsa_pairwise_vdw_grid_prefix","grid");
	  screen_file = parm.query_param("gbsa_pairwise_screen_file","screen.in");
	  tmp = parm.query_param("gbsa_pairwise_solvent_dielectric", "78.300003");
	  sscanf(tmp.c_str(), "%f", &solvent_dielectric);
  }

}

/************************************************/
void GB_Pairwise::initialize(AMBER_TYPER &typer, int argc, char **argv) {

  if(use_score) {
	  cout << "Initializing GB/SA Pairwise Score Routines..." << endl;
	  vdw_score.energy_score = 1;
	  vdw_score.es_scale = 0;
	  vdw_score.vdw_scale = 1;
	  vdw_score.file_prefix = vdw_grid_prefix;
	  vdw_score.initialize(typer);

	  read_gb_receptor_grid();
	  read_sa_receptor_grid();

	  if(check_commandline_flag(argv, argc, "-v2") != -1)
		  verbosity_level = 1;
	  else
		  verbosity_level = 0;
  }
}

/************************************************/
bool GB_Pairwise::compute_score(DOCKMol &mol) {
  float		total;

  if(use_score) {
	  get_gb_solvation_score(mol);
	  get_sa_solvation_score(mol);

	  vdw_score.compute_score(mol);
	  vdw_component = vdw_score.vdw_component;

	  total = gb_component + 0.6*vdw_component -0.015*sa_component + 0.025*del_SAS_hp;

	  mol.current_score = total;

	  return true;
  } else
	  return false;
}

/************************************************/
string GB_Pairwise::output_score_summary(float score) {
  string	text;
  char		line[1000];

  if(use_score) {
	  sprintf(line, "########## GB/SA Score:%21f\n##########%13s%21f\n##########%13s%21f\n##########%13s%21f\n##########%13s%21f\n##########\n", score, "vdw:", vdw_component, "gb:", gb_component, "sa:", sa_component, "del_SAS_hp:", del_SAS_hp);

	  text = line;

	  return text;
  } else
	  return "";

}

/************************************************/
void GB_Pairwise::read_gb_receptor_grid() {
  char 	filename[100], line[500], header[20];
  float	*tmp;
  int	i;
  FILE	*ifp;

  if(use_score) {
	  // read from .bmp file
	  sprintf(filename, "%s.bmp", gb_grid_prefix.c_str());

	  ifp = fopen(filename, "r");

	  if(ifp == NULL) {
		  cout << "\n\nCould not open " << filename << " for reading.  Program will terminate." << endl << endl;
		  exit(0);
	  }

	  fgets(line, 500, ifp);
	  sscanf(line, "%s", header);
	  fgets(line, 500, ifp);
	  sscanf(line, "%f %f %f %f %d %d %d",&gb_grid_spacing,&gb_grid_origin[0],&gb_grid_origin[1],&gb_grid_origin[2],&gb_grid_span[0],&gb_grid_span[1],&gb_grid_span[2]);
	  fgets(line, 500, ifp);
	  sscanf(line, "%f %f %d %f %f %f",&gb_grid_grdcut,&gb_grid_grdcuto,&gb_rec_total_atoms,&gb_grid_cutsq,&gb_grid_f_scale,&gb_rec_solv_rec);

	  fclose(ifp);

	  gb_grid_size = gb_grid_span[0]*gb_grid_span[1]*gb_grid_span[2];

	  // resize receptor vectors
	  gb_rec_charge.clear();
	  gb_rec_coords.clear();
	  gb_rec_inv_a.clear();
	  gb_rec_inv_a_rec.clear();
	  gb_rec_radius_eff.clear();
	  gb_rec_receptor_screen.clear();

	  gb_rec_charge.resize(gb_rec_total_atoms);
	  gb_rec_coords.resize(gb_rec_total_atoms);
	  gb_rec_inv_a.resize(gb_rec_total_atoms);
	  gb_rec_inv_a_rec.resize(gb_rec_total_atoms);
	  gb_rec_radius_eff.resize(gb_rec_total_atoms);
	  gb_rec_receptor_screen.resize(gb_rec_total_atoms);

	  tmp = (float *)malloc((gb_rec_total_atoms+10)*sizeof(float));
/*
	  // section to read from .rec.txt file
	  sprintf(filename, "%s.rec.txt", gb_grid_prefix.c_str());
	  ifp = fopen(filename, "r");

	  if(ifp == NULL) {
		  cout << "\n\nCould not open " << filename << " for reading.  Program will terminate." << endl << endl;
		  exit(0);
	  }

	  for(i=0;i<gb_rec_total_atoms;i++) {
		  fgets(line, 500, ifp);
		  sscanf(line, "%f %f %f %f %f %f %f",&gb_rec_coords[i].crd[0], &gb_rec_coords[i].crd[1], &gb_rec_coords[i].crd[2], &gb_rec_charge[i], &gb_rec_radius_eff[i], &gb_rec_inv_a_rec[i], &gb_rec_receptor_screen[i]);
	  }

	  fclose(ifp);
*/
	  // read from binary .rec file instead of .rec.txt
	  sprintf(filename, "%s.rec", gb_grid_prefix.c_str());
	  ifp = fopen(filename, "rb");

	  if(ifp == NULL) {
		  cout << "\n\nCould not open " << filename << " for reading.  Program will terminate." << endl << endl;
		  exit(0);
	  }


	  fread(tmp, sizeof(float), gb_rec_total_atoms+2, ifp);
	  for(i=1;i<=gb_rec_total_atoms;i++)
		  gb_rec_coords[i-1].crd[0] = tmp[i];

	  fread(tmp, sizeof(float), gb_rec_total_atoms+2, ifp);
	  for(i=1;i<=gb_rec_total_atoms;i++)
		  gb_rec_coords[i-1].crd[1] = tmp[i];

	  fread(tmp, sizeof(float), gb_rec_total_atoms+2, ifp);
	  for(i=1;i<=gb_rec_total_atoms;i++)
		  gb_rec_coords[i-1].crd[2] = tmp[i];

	  fread(tmp, sizeof(float), gb_rec_total_atoms+2, ifp);
	  for(i=1;i<=gb_rec_total_atoms;i++)
		  gb_rec_charge[i-1] = tmp[i];

	  fread(tmp, sizeof(float), gb_rec_total_atoms+2, ifp);
	  for(i=1;i<=gb_rec_total_atoms;i++)
		  gb_rec_radius_eff[i-1] = tmp[i];

	  fread(tmp, sizeof(float), gb_rec_total_atoms+2, ifp);
	  for(i=1;i<=gb_rec_total_atoms;i++)
		  gb_rec_inv_a_rec[i-1] = tmp[i];

	  fread(tmp, sizeof(float), gb_rec_total_atoms+2, ifp);
	  for(i=1;i<=gb_rec_total_atoms;i++)
		  gb_rec_receptor_screen[i-1] = tmp[i];

	  fclose(ifp);
	  free(tmp);

	  gb_initialized = false;
  }
}

/************************************************/
float GB_Pairwise::get_gb_solvation_score(DOCKMol &mol) {
  float	factor, sub_total, solv_lp, es, temp;
  float	crd[3];
  float rvdweff0, rvdweff2, sum;
  float r2, rij2, inv_aij2, dij, f_gb;
  int	nearpt[3];
  int	i, j;
  float INVA_NEG = (float)0.1;
  float ri,sj,sj2,rij,RLij,RUij,RLij2,RUij2,tmp_pairwise;
  float screen[25];
  float	ligand_solvation, ligand_solvation_partial;
  float	receptor_solvation, receptor_solvation_partial;
  float	ligand_desolvation, receptor_desolvation, screened_es;
  FILE *fp1;

  if(use_score) {
	  gb_initialized = false;

	  // read in the screen.in file
	  fp1=fopen(screen_file.c_str(),"r");

	  if(fp1 == NULL) {
		  cout << "\n\nCould not open " << screen_file << " for reading.  Program will terminate." << endl << endl;
		  exit(0);
	  }

	  for (i=0;i<25;i++) {
	    fscanf(fp1,"%f",&screen[i]);
	  }
	  fclose(fp1);

	  // set the solvent dielectric factor (hardcoded for now)
	  factor = - 166.0 * (1.0-1.0/solvent_dielectric);
	  sub_total = 0.0;

	  // Conditional- initialize GB if necessary
	  if(!gb_initialized) {
		  gb_initialized = true;

		  // Clear & resize ligand GB arrays
		  gb_lig_coords.clear();
		  gb_lig_inv_a.clear();
		  gb_lig_radius_eff.clear();
		  gb_lig_sum_lig.clear();
		  gb_lig_flag.clear();

		  gb_lig_coords.resize(mol.num_atoms);
		  gb_lig_inv_a.resize(mol.num_atoms);
		  gb_lig_radius_eff.resize(mol.num_atoms);
		  gb_lig_sum_lig.resize(mol.num_atoms);
		  gb_lig_flag.resize(mol.num_atoms);

		  // Initialize ligand GB arrays
		  for(i=0;i<mol.num_atoms;i++) {
			  gb_lig_coords[i].crd[0] = 0.0;
			  gb_lig_coords[i].crd[1] = 0.0;
			  gb_lig_coords[i].crd[2] = 0.0;
			  gb_lig_inv_a[i] = 0.0;
			  gb_lig_radius_eff[i] = 0.0;
			  gb_lig_sum_lig[i] = 0.0;
			  gb_lig_flag[i] = 0;
		  }

		  // ID ligand atoms->grid points
		  for(i=0;i<mol.num_atoms;i++) {
			  if(mol.amber_at_well_depth[i] != 0.0) {

				  // Label atoms that fall outside the grid
				  crd[0] = mol.x[i] - gb_grid_origin[0];
				  crd[1] = mol.y[i] - gb_grid_origin[1];
				  crd[2] = mol.z[i] - gb_grid_origin[2];

				  for(j=0;j<3;j++) {
					  nearpt[j] = NINT(crd[j]/gb_grid_spacing);
					  if((nearpt[j] < 0) | (nearpt[j] >= gb_grid_span[j])) {
						  gb_lig_flag[i] = 1;
					  }
					  gb_lig_coords[i].crd[j] = crd[j];
				  }

				  rvdweff0 = mol.amber_at_radius[i] + VDWOFF;
				  rvdweff2 = rvdweff0*rvdweff0;
				  gb_lig_radius_eff[i] = rvdweff0;

			  }
		  }

		  // Find solvation energy for pure ligand
		  for(i=0;(i<mol.num_atoms)&&(gb_lig_flag[i]==0);i++) {
			  if(mol.amber_at_well_depth[i] != 0.0) {

				  rvdweff0 = gb_lig_radius_eff[i];
				  rvdweff2 = rvdweff0*rvdweff0;
				  sum = 0.0;
				  ri=rvdweff0-VDWOFF;

				  // Pairwise calculation of Born Radius
				  for(j=0;(j<mol.num_atoms)&&(gb_lig_flag[j]==0)&&(j!=i);j++) {
					  if(mol.amber_at_well_depth[j] != 0.0) {
						  rvdweff0 = gb_lig_radius_eff[j];
						  sj = (rvdweff0-VDWOFF)*screen[mol.amber_at_id[j]+1];
						  sj2=sj*sj;
						  r2 = square_distance (gb_lig_coords[i].crd, gb_lig_coords[j].crd);
						  rij=sqrt(r2);
						  if (ri<rij+sj) {
							  RLij=1./MAX(ri,rij-sj);
							  RUij=1./(rij+sj);
							  RLij2=RLij*RLij;
							  RUij2=RUij*RUij;
							  tmp_pairwise=RLij-RUij+0.25*(sj2/rij-rij)*(RLij2-RUij2)-0.5/rij*log(RLij/RUij);
						  } else {
							  tmp_pairwise=0;
						  }

						  sum+=tmp_pairwise;
					  }
				  } // End pairwise born radii

				  gb_lig_sum_lig[i] = sum;
				  gb_lig_inv_a[i] = 1.0/(ri) - 0.5*sum;
				  if(gb_lig_inv_a[i] < 0) {
					  cout << "Pure ligand inv_a error: " << i << "\t" << gb_lig_inv_a[i] << endl;
					  gb_lig_inv_a[i] = INVA_NEG;
				  }
			  }
		  }

		  ligand_solvation = 0.0;

		  for(i=0;(i<mol.num_atoms)&&(gb_lig_flag[i]==0);i++) {
			  for(j=0;(j<mol.num_atoms)&&(gb_lig_flag[j]==0);j++) {
				  rij2 = square_distance (gb_lig_coords[i].crd, gb_lig_coords[j].crd);
				  if(rij2 <= gb_grid_cutsq) {
					  inv_aij2 = gb_lig_inv_a[i] * gb_lig_inv_a[j];
					  dij = rij2 * inv_aij2 / gb_grid_f_scale;
					  f_gb = sqrt(rij2 + exp(-dij)/inv_aij2);
					  ligand_solvation += factor * mol.charges[i] * mol.charges[j] / f_gb;
				  }
			  }
		  } // End Ligand solvation energy

		  // Calculate effect of receptor atoms on ligand atom Born radii
		  for(i=0;(i<mol.num_atoms)&&(gb_lig_flag[i]==0);i++) {
			  if(mol.amber_at_well_depth[i] != 0.0) {
				  rvdweff0 = gb_lig_radius_eff[i];
				  sum = gb_lig_sum_lig[i];
				  ri=rvdweff0-VDWOFF;

				  for(j=0;j<gb_rec_total_atoms;j++) {
					  rvdweff0 = gb_rec_radius_eff[j];
					  sj = (rvdweff0-VDWOFF)*gb_rec_receptor_screen[j];
					  sj2 = sj*sj;
					  r2 = square_distance(gb_lig_coords[i].crd, gb_rec_coords[j].crd);
					  rij = sqrt(r2);
					  if (ri<rij+sj) {
						  RLij=1./MAX(ri,rij-sj);
						  RUij=1./(rij+sj);
						  RLij2=RLij*RLij;
						  RUij2=RUij*RUij;
						  tmp_pairwise=RLij-RUij+0.25*(sj2/rij-rij)*(RLij2-RUij2)-0.5/rij*log(RLij/RUij);
					  } else {
						  tmp_pairwise=0;
					  }

					  sum+=tmp_pairwise;
				  } // End receptor loop

				  gb_lig_inv_a[i] = 1.0/(ri) - 0.5*sum;

				  if(gb_lig_inv_a[i] < 0) {
					  cout << "Receptor on ligand inv_a error: " << i << "\t" << gb_lig_inv_a[i] << endl;
					  gb_lig_inv_a[i] = INVA_NEG;
				  }
			  }
		  } // End receptor atom -> ligand born radii calculation

		  // Calculate effect of ligand atoms on receptor atom born radii
		  for(i=0;i<gb_rec_total_atoms;i++) {
			  rvdweff0 = gb_rec_radius_eff[i];
			  sum = 0.0;

			  ri=rvdweff0-VDWOFF;
			  for(j=0;(j<mol.num_atoms)&&(gb_lig_flag[j]==0);j++) {
				  rvdweff0 = gb_lig_radius_eff[j];
				  sj = (rvdweff0-VDWOFF)*screen[mol.amber_at_id[j]+1];
				  sj2=sj*sj;
				  r2 = square_distance(gb_rec_coords[i].crd, gb_lig_coords[j].crd);
				  rij=sqrt(r2);

				  if (ri<rij+sj) {
					  RLij=1./MAX(ri,rij-sj);
					  RUij=1./(rij+sj);
					  RLij2=RLij*RLij;
					  RUij2=RUij*RUij;
					  tmp_pairwise=RLij-RUij+0.25*(sj2/rij-rij)*(RLij2-RUij2)-0.5/rij*log(RLij/RUij);
				  } else {
					  tmp_pairwise=0;
				  }

				  sum+=tmp_pairwise;
			  }

			  gb_rec_inv_a[i] = gb_rec_inv_a_rec[i] - sum * 0.5;

			  if(gb_rec_inv_a[i] < 0) {
				  cout << "Ligand on receptor inva error: " << i << "\t" << gb_rec_inv_a[i] << endl;
				  gb_rec_inv_a[i] = INVA_NEG;
			  }

		  } // End ligand atom -> receptor born radii calculation

		  // Polarization solvation energy of L-P complex
		  // ll term in solv_complex
		  ligand_solvation_partial = 0.0;
		  for(i=0;(i<mol.num_atoms)&&(gb_lig_flag[i]==0);i++) {
			  for(j=0;(j<mol.num_atoms)&&(gb_lig_flag[j]==0);j++) {
				  rij2 = square_distance(gb_lig_coords[i].crd, gb_lig_coords[j].crd);

				  if(rij2 <= gb_grid_cutsq) {
					  inv_aij2 = gb_lig_inv_a[i] * gb_lig_inv_a[j];
					  dij = rij2 * inv_aij2 / gb_grid_f_scale;
					  f_gb = sqrt(rij2 + exp(-dij)/inv_aij2);
					  ligand_solvation_partial += factor * mol.charges[i] * mol.charges[j] / f_gb;
				  }
			  }
		  }

		  // pp term in solv_complex
		  receptor_solvation_partial = 0.0;
		  for(i=0;i<gb_rec_total_atoms;i++) {
			  for(j=0;j<gb_rec_total_atoms;j++) {
				  rij2 = square_distance(gb_rec_coords[i].crd, gb_rec_coords[j].crd);

				  if(rij2 <= gb_grid_cutsq) {
					  inv_aij2 = gb_rec_inv_a[i] * gb_rec_inv_a[j];
					  dij = rij2 * inv_aij2 / gb_grid_f_scale;
					  f_gb = sqrt(rij2 + exp(-dij)/inv_aij2);

					  receptor_solvation_partial += factor * gb_rec_charge[i] * gb_rec_charge[j] / f_gb;

				  }
			  }
		  }

		  receptor_solvation = gb_rec_solv_rec;

	  } else { // end initialization routine

		  for(i=0;(i<mol.num_atoms)&&(gb_lig_flag[i]==0);i++) {
			  if(mol.amber_at_well_depth[i] != 0.0) {
				  for(j=0;j<3;j++) {
					  crd[j] = gb_lig_coords[i].crd[j] - gb_grid_origin[j];
					  nearpt[j]=NINT(crd[j]/gb_grid_spacing);

					  if ((nearpt[j] < 0) | (nearpt[j] >= gb_grid_span[j])) {
						  gb_lig_flag[i] = 1;
					  }

					  gb_lig_coords[i].crd[j] = crd[j];
				  }
			  }
		  }

	  } // end all initialization routines

	  // find the atomic contribution to screened es interactions
	  solv_lp = 0.0;
	  es = 0.0;

	  for(i=0;(i<mol.num_atoms)&&(gb_lig_flag[i]==0);i++) {
		  temp = 0.0;

		  for(j=0;j<gb_rec_total_atoms;j++) {
			  rij2 = square_distance (gb_lig_coords[i].crd, gb_rec_coords[j].crd);

			  if(rij2<gb_grid_cutsq) {
				  inv_aij2 = gb_lig_inv_a[i] * gb_rec_inv_a[j];
				  dij = rij2 * inv_aij2 / gb_grid_f_scale;
				  f_gb = sqrt(rij2 + exp(-dij)/inv_aij2);
				  solv_lp += 2 * factor * mol.charges[i] * gb_rec_charge[j] / f_gb;
				  es += 332 * mol.charges[i] * gb_rec_charge[j] / sqrt(rij2);
				  temp += 2 *factor* mol.charges[i] * gb_rec_charge[j] / f_gb+332*mol.charges[i] * gb_rec_charge[j] / sqrt(rij2);
			  }
		  }
	  }

	  ligand_desolvation = ligand_solvation_partial - ligand_solvation;
	  receptor_desolvation = receptor_solvation_partial - receptor_solvation;
	  screened_es = solv_lp + es;
	  gb_component = ligand_desolvation + receptor_desolvation + screened_es;

	  if(verbosity_level == 1) {
		  cout << "\nGB Calculation" << endl;
		  cout << "--------------------" << endl;
		  cout << "ligand desolvation:\t\t" << ligand_desolvation << endl;
		  cout << "receptor desolvation:\t\t" << receptor_desolvation << endl;
		  cout << "screened electrostatics:\t" << screened_es << endl;
		  cout << "total:\t\t\t\t" << gb_component << endl;;
	  }

	  return gb_component;
  } else
	  return 0;

}

/************************************************/
void GB_Pairwise::read_sa_receptor_grid() {
  char 	filename[100], line[500], header[20];
  int	i, j;
  float	*tmp;
  int	*tmp_int;
  FILE	*ifp;

  if(use_score) {
	  // read the .bmp file
	  sprintf(filename, "%s.bmp", sa_grid_prefix.c_str());

	  ifp = fopen(filename, "r");

	  if(ifp == NULL) {
		  cout << "\n\nCould not open " << filename << " for reading.  Program will terminate." << endl << endl;
		  exit(0);
	  }

	  fgets(line, 500, ifp);
	  sscanf(line, "%s", header);
	  fgets(line, 500, ifp);
	  sscanf(line, "%f %f %f %f %d %d %d",&sa_grid_spacing,&sa_grid_origin[0],&sa_grid_origin[1],&sa_grid_origin[2],&sa_grid_span[0],&sa_grid_span[1],&sa_grid_span[2]);
	  fgets(line, 500, ifp);
	  sscanf(line, "%f %f %d %f %f %f %d",&sa_grid_grdcut,&sa_grid_grdcuto,&sa_rec_total_atoms,&sa_grid_cutsq,&sa_grid_f_scale,&sa_rec_solv_rec,&sa_grid_nvtyp);

	  fclose(ifp);

	  sa_grid_size = sa_grid_span[0]*sa_grid_span[1]*sa_grid_span[2];

	  // resize receptor vectors
	  sa_rec_coords.clear();
	  sa_rec_radius_eff.clear();
	  sa_rec_vdwn_rec.clear();
	  sa_rec_nsphgrid.clear();
	  sa_rec_nhp.clear();

	  sa_rec_coords.resize(sa_rec_total_atoms);
	  sa_rec_radius_eff.resize(sa_rec_total_atoms);
	  sa_rec_vdwn_rec.resize(sa_rec_total_atoms);
	  sa_rec_nsphgrid.resize(sa_grid_nvtyp);
	  sa_rec_nhp.resize(sa_grid_nvtyp);

	  /***
	  // read the .sas.txt file
	  sprintf(filename, "%s.sas.txt", sa_grid_prefix.c_str());
	  ifp = fopen(filename, "r");

	  if(ifp == NULL) {
		  cout << "\n\nCould not open " << filename << " for reading.  Program will terminate." << endl << endl;
		  exit(0);
	  }

	  for(i=0;i<sa_rec_total_atoms;i++) {
		  fgets(line, 500, ifp);
		  sscanf(line, "%f", &sa_rec_coords[i].crd[0]);
	  }
	  for(i=0;i<sa_rec_total_atoms;i++) {
		  fgets(line, 500, ifp);
		  sscanf(line, "%f", &sa_rec_coords[i].crd[1]);
	  }
	  for(i=0;i<sa_rec_total_atoms;i++) {
		  fgets(line, 500, ifp);
		  sscanf(line, "%f", &sa_rec_coords[i].crd[2]);
	  }
	  for(i=0;i<sa_rec_total_atoms;i++) {
		  fgets(line, 500, ifp);
		  sscanf(line, "%f", &sa_rec_radius_eff[i]);
	  }
	  for(i=0;i<sa_rec_total_atoms;i++) {
		  fgets(line, 500, ifp);
		  sscanf(line, "%d", &sa_rec_vdwn_rec[i]);
	  }
	  for(i=0;i<sa_grid_nvtyp;i++) {
		  fgets(line, 500, ifp);
		  sscanf(line, "%d", &sa_rec_nsphgrid[i]);
	  }
	  for(i=0;i<sa_grid_nvtyp;i++) {
		  fgets(line, 500, ifp);
		  sscanf(line, "%d", &sa_rec_nhp[i]);
	  }

	  fclose(ifp);
	  ***/

	  tmp = (float *)malloc((sa_rec_total_atoms+10)*sizeof(float));
	  tmp_int = (int *)malloc((sa_rec_total_atoms+10)*sizeof(int));

	  // read the .sas binary file
	  sprintf(filename, "%s.sas", sa_grid_prefix.c_str());
	  ifp = fopen(filename, "rb");

	  if(ifp == NULL) {
		  cout << "\n\nCould not open " << filename << " for reading.  Program will terminate." << endl << endl;
		  exit(0);
	  }

	  fread(tmp, sizeof(float), sa_rec_total_atoms+2, ifp);
	  for(i=0;i<sa_rec_total_atoms;i++) {
		  sa_rec_coords[i].crd[0] = tmp[i+1];
	  }

	  fread(tmp, sizeof(float), sa_rec_total_atoms+2, ifp);
	  for(i=0;i<sa_rec_total_atoms;i++) {
		  sa_rec_coords[i].crd[1] = tmp[i+1];
	  }

	  fread(tmp, sizeof(float), sa_rec_total_atoms+2, ifp);
	  for(i=0;i<sa_rec_total_atoms;i++) {
		  sa_rec_coords[i].crd[2] = tmp[i+1];
	  }

	  fread(tmp, sizeof(float), sa_rec_total_atoms+2, ifp);
	  for(i=0;i<sa_rec_total_atoms;i++) {
		  sa_rec_radius_eff[i] = tmp[i+1];
	  }

	  fread(tmp_int, sizeof(int), sa_rec_total_atoms+2, ifp);
	  for(i=0;i<sa_rec_total_atoms;i++) {
		  sa_rec_vdwn_rec[i] = tmp_int[i+1];
	  }

	  fread(tmp_int, sizeof(int), sa_grid_nvtyp+2, ifp);
	  for(i=0;i<sa_grid_nvtyp;i++) {
		  sa_rec_nsphgrid[i] = tmp_int[i+1];
	  }

	  fread(tmp_int, sizeof(int), sa_grid_nvtyp+2, ifp);
	  for(i=0;i<sa_grid_nvtyp;i++) {
		  sa_rec_nhp[i] = tmp_int[i+1];
	  }

	  fclose(ifp);
	  free(tmp);
	  free(tmp_int);

	  sa_grid_sphgrid_crd.resize(sa_grid_nvtyp);
	  for(i=0;i<sa_grid_nvtyp;i++) {
		  sa_grid_sphgrid_crd[i].resize(sa_rec_nsphgrid[i]);
	  }

	  sa_grid_mark_sas.resize(sa_rec_total_atoms);
	  for(i=0;i<sa_rec_total_atoms;i++) {
		  sa_grid_mark_sas[i].resize(sa_rec_nsphgrid[sa_rec_vdwn_rec[i]-1]);
	  }

	  // read the .sasmark file
	  sprintf(filename, "%s.sasmark", sa_grid_prefix.c_str());
	  ifp = fopen(filename, "r");

	  if(ifp == NULL) {
		  cout << "\n\nCould not open " << filename << " for reading.  Program will terminate." << endl << endl;
		  exit(0);
	  }

	  fgets(line, 500, ifp);
	  sscanf(line, "%d %d %d", &sa_rec_nsas, &sa_rec_nsas_hp, &sa_rec_nsas_pol);
	  fgets(line, 500, ifp);
	  sscanf(line, "%f %f %f", &sa_grid_r_probe, &sa_grid_sspacing, &sa_grid_r2_cutoff);

	  for(i=0;i<sa_grid_nvtyp;i++) {
		  for(j=0;j<sa_rec_nsphgrid[i];j++) {
			  fscanf(ifp, "%f %f %f", &sa_grid_sphgrid_crd[i][j].crd[0], &sa_grid_sphgrid_crd[i][j].crd[1], &sa_grid_sphgrid_crd[i][j].crd[2]);
		  }
	  }

	  for(i=0;i<sa_rec_total_atoms;i++) {
		  for(j=0;j<sa_rec_nsphgrid[sa_rec_vdwn_rec[i]-1];j++) {
			  fscanf(ifp, "%d", &sa_grid_mark_sas[i][j]);
		  }
	  }

	  fclose(ifp);
  }

}

/************************************************/
float GB_Pairwise::get_sa_solvation_score(DOCKMol &mol) {
  int	i, j, k, l, n, i1, mark;
  float	crd[3], crd0[3], crd1[3], r2;
  int	vdwn, nsum;
  int	nearpt[3];
  int	neighbor_num[MAX_ATOM_LIG], neighbors[MAX_ATOM_LIG][MAX_ATOM_LIG];
  int	neighbor_num1[MAX_ATOM_REC], neighbors1[MAX_ATOM_REC][MAX_ATOM_LIG];
  int	nsas_lig, delta_nsas_rec, delta_nsas_lig;
  int	nsas_hp_lig, delta_nsas_hp_rec, delta_nsas_hp_lig;
  int	nsas_pol_lig, delta_nsas_pol_rec, delta_nsas_pol_lig;
  float	score[2], factor;

  if(use_score) {
	  // clear all ligand parameters
	  sa_grid_mark_sas_lig.clear();
	  sa_lig_coords.clear();
	  sa_lig_radius_eff.clear();

	  // resize all ligand parameters
	  sa_grid_mark_sas_lig.resize(mol.num_atoms);
	  sa_lig_coords.resize(mol.num_atoms);
	  sa_lig_radius_eff.resize(mol.num_atoms);

	  // initialize ligand parameters
	  for(i=0;i<mol.num_atoms;i++)
		  for(j=0;j<2000;j++)
			  sa_grid_mark_sas_lig[i].vals[j] = 0;


	  for(i=0;i<mol.num_atoms;i++) {
		  sa_lig_coords[i].crd[0] = 0.0;
		  sa_lig_coords[i].crd[1] = 0.0;
		  sa_lig_coords[i].crd[2] = 0.0;
		  sa_lig_radius_eff[i] = 0.0;
		  neighbor_num[i]=0;
		  neighbor_num1[i]=0;
		  for (j = 0; j < 150; j++) {
			  neighbors[i][j]=0;
			  neighbors1[i][j]=0;
		  }
	  }

	  for(i=0;i<sa_rec_total_atoms;i++) {
		  n = i + mol.num_atoms;
		  neighbor_num1[n] = 0;
		  for(j=0;j<150;j++)
			  neighbors1[n][j] = 0;
	  }

	  // Pure ligand atoms
	  for(i=0;i<mol.num_atoms;i++) {
		  if(mol.amber_at_well_depth[i] != 0.0) {

			  crd[0] = mol.x[i] - sa_grid_origin[0];
			  crd[1] = mol.y[i] - sa_grid_origin[1];
			  crd[2] = mol.z[i] - sa_grid_origin[2];

			  for(j=0;j<3;j++) {
				  nearpt[j] = NINT(crd[j]/sa_grid_spacing);

//				  if ((nearpt[j] < 0) | (nearpt[j] >= sa_grid_span[j])) {
//					  cout << "Ligand atoms out of box!" << endl;
//				  }

				  sa_lig_coords[i].crd[j] = crd[j];
			  }

			  r2 = SQR(mol.amber_at_radius[i] + sa_grid_r_probe);
			  sa_lig_radius_eff[i] = r2;
		  }
	  }

	  // find neighboring atoms
	  for(i=0;i<mol.num_atoms;i++) {
		  if(mol.amber_at_well_depth[i] != 0.0) {

			  for(k=0;k<3;k++)
				  crd[k] = sa_lig_coords[i].crd[k];

			  for(j=i+1;j<mol.num_atoms;j++) {
				  if(mol.amber_at_well_depth[j] != 0.0) {

					  for(k=0;k<3;k++)
						  crd1[k] = sa_lig_coords[j].crd[k];

					  r2 = square_distance(crd, crd1);
					  if (r2 < sa_grid_r2_cutoff)  {
						  neighbor_num[i] ++;
						  neighbor_num[j] ++;
						  neighbors[i][neighbor_num[i]]=j;
						  neighbors[j][neighbor_num[j]]=i;
					  }
				  }
			  }

			  for(j=0;j<sa_rec_total_atoms;j++) {
				  for(k=0;k<3;k++)
					  crd1[k] = sa_rec_coords[j].crd[k];

				  n = j + mol.num_atoms;
				  r2 = square_distance(crd, crd1);
				  if(r2 < sa_grid_r2_cutoff) {
					  neighbor_num1[i] ++;
					  neighbor_num1[n] ++;
					  neighbors1[i][neighbor_num1[i]]=n;
					  neighbors1[n][neighbor_num1[n]]=i;
				  }
			  }

		  }
	  }

	  // Find SASA for pure ligand
	  nsas_lig = 0;
	  nsas_hp_lig = 0;
	  nsas_pol_lig = 0;

	  for(i=0;i<mol.num_atoms;i++) {
		  if(mol.amber_at_well_depth[i] != 0.0) {
			  vdwn = mol.amber_at_id[i];
			  nsum = 0;

			  for(k=0;k<3;k++)
				  crd0[k] = sa_lig_coords[i].crd[k];

			  for(l=0;l<sa_rec_nsphgrid[vdwn];l++) {
				  for(k=0;k<3;k++)
					  crd[k] = crd0[k] + sa_grid_sphgrid_crd[vdwn][l].crd[k];

				  sa_grid_mark_sas_lig[i].vals[l] = 0;
				  mark = 0;

				  for(j=1;j<=neighbor_num[i];j++) {
					  n = neighbors[i][j];

					  for(k=0;k<3;k++)
						  crd1[k] = sa_lig_coords[n].crd[k];

					  r2 = square_distance(crd, crd1);
					  if(r2 < sa_lig_radius_eff[n]){
						  mark = 1;
						  break;
					  }
				  }

				  if(mark == 0) {
					  sa_grid_mark_sas_lig[i].vals[l] = 1;
					  nsas_lig++;
					  nsum++;
				  }
			  }

			  if(sa_rec_nhp[vdwn] == 1)
				  nsas_hp_lig += nsum;
			  else
				  nsas_pol_lig += nsum;
		  }
	  }


	  // Effect of receptor atoms on SAS of ligand atoms
	  delta_nsas_lig = 0;
	  delta_nsas_hp_lig = 0;
	  delta_nsas_pol_lig = 0;

	  for(i=0;i<mol.num_atoms;i++) {
		  if(mol.amber_at_well_depth[i] != 0.0) {
			  nsum = 0;
			  vdwn = mol.amber_at_id[i];

			  for (k=0;k<3;k++)
				  crd0[k] = sa_lig_coords[i].crd[k];

			  for(l=0;l<sa_rec_nsphgrid[vdwn];l++) {
				  if(sa_grid_mark_sas_lig[i].vals[l] == 1) {

					  for(k=0;k<3;k++)
						  crd[k] = crd0[k] + sa_grid_sphgrid_crd[vdwn][l].crd[k];

					  for (j=1;j<=neighbor_num1[i];j++) {
						  n=neighbors1[i][j] - mol.num_atoms;

						  for(k=0;k<3;k++)
							  crd1[k] = sa_rec_coords[n].crd[k];

						  r2 = square_distance(crd, crd1);
						  if(r2 < sa_rec_radius_eff[n]) {
							  delta_nsas_lig ++;
							  nsum++;
							  break;
						  }
					  }
				  }
			  }

			  if(sa_rec_nhp[vdwn] == 1)
				  delta_nsas_hp_lig += nsum;
			  else
				  delta_nsas_pol_lig += nsum;
		  }
	  }

	  // Effect of ligand atoms on SAS of receptor atoms
	  delta_nsas_rec = 0;
	  delta_nsas_hp_rec = 0;
	  delta_nsas_pol_rec = 0;

	  for(i=0;i<sa_rec_total_atoms;i++) {
		  nsum = 0;
		  vdwn = sa_rec_vdwn_rec[i] - 1;

		  for(k=0;k<3;k++)
			  crd0[k] = sa_rec_coords[i].crd[k];

		  i1 = i + mol.num_atoms;

		  for(l=0;l<sa_rec_nsphgrid[vdwn];l++) {
			  if(sa_grid_mark_sas[i][l] == 1) {
				  for(k=0;k<3;k++)
					  crd[k] = crd0[k] + sa_grid_sphgrid_crd[vdwn][l].crd[k];

				  for (j=1;j<=neighbor_num1[i1];j++) {
					  n=neighbors1[i1][j];

					  for (k=0; k<3; k++)
						  crd1[k] = sa_lig_coords[n].crd[k];

					  r2 = square_distance (crd, crd1);
					  if (r2 < sa_lig_radius_eff[n]) {
						  delta_nsas_rec ++;
						  nsum ++;
						  break;
					  }

				  }
			  }
		  }

		  if (sa_rec_nhp[vdwn] == 1)
	        delta_nsas_hp_rec += nsum;
	      else
	        delta_nsas_pol_rec += nsum;
	  }

	  factor = SQR(sa_grid_sspacing);
	  score[0] = - (float) (delta_nsas_rec) * factor;
	  score[1] = - (float) (delta_nsas_lig) * factor;

	  sa_component = score[0] + score[1];
	  del_SAS_hp = (- (delta_nsas_hp_rec + delta_nsas_hp_lig) * factor);

	  if(verbosity_level == 1) {
		  cout << "\nSA Calculation:" << endl;
		  cout << "--------------------" << endl;
		  cout << "SAS_rec = " << (sa_rec_nsas * factor);
		  cout << ", delta_SAS_rec = " << score[0] << endl;

		  cout << "SAS_lig = " << (nsas_lig * factor);
		  cout << ", delta_SAS_lig = " << score[1] << endl;

		  cout << "delta_SAS_pol = " << (- (delta_nsas_pol_rec + delta_nsas_pol_lig) * factor);
		  cout << ", delta_SAS_hp = " << (- (delta_nsas_hp_rec + delta_nsas_hp_lig) * factor);
		  cout << endl << "total = " << (score[0] + score[1]) << endl << endl;
	  }

	  return sa_component;
  } else
	  return 0;

}

/************************************************/
/************************************************/
/************************************************/
float born_radii_calc(float Rkkp, float Pkp, float Pk) {
	float	Lkkp, Ukkp, iLkkp, iLkkp2, iUkkp, iUkkp2;
	float	sum1, sum2, sum3, sum4;

	// Calculate Lkkp
	Lkkp = 0.0;

	if((Rkkp + Pkp) <= Pk)
		Lkkp = 1.0;
	if((Rkkp - Pkp) <= Pk)
		Lkkp = Pk;
	if(Pk < (Rkkp + Pkp))
		Lkkp = Pk;
	if(Pk <= (Rkkp - Pkp))
		Lkkp = (Rkkp - Pkp);

	// calculate Ukkp
	Ukkp = 0.0;

	if((Rkkp + Pkp) <= Pk)
		Ukkp = 1.0;
	if(Pk < (Rkkp + Pkp))
		Ukkp = (Rkkp + Pkp);

	// compute some inverse numbers
	iLkkp = 1.0/Lkkp;
	iLkkp2 = pow(iLkkp, 2);

	iUkkp = 1.0/Ukkp;
	iUkkp2 = pow(iUkkp, 2);

	sum1 = iLkkp - iUkkp;
	sum2 = ((Rkkp/4.0)*(iUkkp2-iLkkp2));
	sum3 = ((1.0/(2.0*Rkkp))*(log(Lkkp/Ukkp)));
	sum4 = (((Pkp*Pkp)/(4.0*Rkkp))*(iLkkp2-iUkkp2));

	return sum1 + sum2 + sum3 + sum4;
}

/************************************************/
void GB_Hawkins::input_parameters(Parameter_Reader &parm, bool &primary_score, bool &secondary_score) {
  string tmp;

  use_primary_score = false;
  use_secondary_score = false;

  cout << "\nGB/SA Cramer/Truhlar/Hawkins Score Parameters" << endl;
  cout << "------------------------------------------------------------------------------------------" << endl;

  if(!primary_score) {
	  tmp = parm.query_param("gbsa_cth_score_primary","no","yes no");
	  if(tmp == "yes")
		  use_primary_score = true;
	  else
		  use_primary_score = false;

	  primary_score = use_primary_score;
  }

  if(!secondary_score) {
	  tmp = parm.query_param("gbsa_cth_score_secondary","no","yes no");
	  if(tmp == "yes")
		  use_secondary_score = true;
	  else
		  use_secondary_score = false;

	  secondary_score = use_secondary_score;
  }

  if(use_primary_score || use_secondary_score)
	  use_score = true;
  else
	  use_score = false;

  if(use_score) {
	  rec_filename = parm.query_param("gbsa_cth_score_rec_filename", "receptor.mol2");

	  gb_offset = atof(parm.query_param("gbsa_cth_score_gb_offset", "0.09").c_str());
	  vdw_att_exp = atoi(parm.query_param("gbsa_cth_score_vdw_att_exp", "6").c_str());
	  vdw_rep_exp = atoi(parm.query_param("gbsa_cth_score_vdw_rep_exp", "12").c_str());

  }

}

/************************************************/
void GB_Hawkins::initialize(AMBER_TYPER &typer) {
	ifstream		rec_file;

	use_score = true;

	if(use_score) {
		cout << "Initializing GB/SA Cramer/Truhlar/Hawkins Score Routines..." << endl;
		////////////// Hardcoded -> param file
		/*
		gb_offset = 0.09;
		vdw_att_exp = 6;
		vdw_rep_exp = 12;
		rec_filename = "receptor.mol2";
		*/

		// initialize GB members
		brad_rec_assigned = false;
		brad_lig_assigned = false;
		brad_cmp_assigned = false;
		rec_dist_assigned = false;

		// initialize VDW members
		vdwA_assigned = false;
		vdwB_assigned = false;

		init_vdw_energy(typer);

		// read in receptor
		rec_file.open(rec_filename.c_str());

		if(rec_file.fail()) {
			cout << "Error Opening Receptor File!" << endl;
			exit(0);
		}

		if(!Read_Mol2(receptor, rec_file)) {
			cout << "Error Reading Receptor Molecule!" << endl;
			exit(0);
		}

		rec_file.close();

		typer.prepare_molecule(receptor);

		// calculate receptor alone
		prepare_receptor();
	}
}

/************************************************/
void GB_Hawkins::init_vdw_energy(AMBER_TYPER &typer) {
  int		i;

  if(vdwA_assigned) {
	  delete [] vdwA;
	  vdwA_assigned = false;
  }

  if(vdwB_assigned) {
	  delete [] vdwB;
	  vdwB_assigned = false;
  }

  vdwA = new float[typer.d4_atom_typer.types.size()];
  vdwA_assigned = true;
  vdwB = new float[typer.d4_atom_typer.types.size()];
  vdwB_assigned = true;

  for(i=0;i<typer.d4_atom_typer.types.size();i++) {
	  vdwA[i] = sqrt( typer.d4_atom_typer.types[i].well_depth * pow((2*typer.d4_atom_typer.types[i].radius), vdw_rep_exp) );
	  vdwB[i] = sqrt( 2*typer.d4_atom_typer.types[i].well_depth * pow((2*typer.d4_atom_typer.types[i].radius), vdw_att_exp) );
  }

}

/************************************************/
bool GB_Hawkins::compute_score(DOCKMol &mol) {

	if(use_score) {

		prepare_ligand(mol);
		prepare_complex(mol);

		total_score = vdw_component + es_component + gpol_cmp - (gpol_rec + gpol_lig);
		current_score = total_score;
/*
		cout << "\n\nVDW Value:\t\t" << vdw_component << endl;
		cout << "ES  Value:\t\t" << es_component << endl;
		cout << "GB  Value:\t\t" << (gpol_cmp - (gpol_rec + gpol_lig)) << endl;
		cout << " Cmp:\t\t\t" << gpol_cmp << "\n Rec:\t\t\t" << gpol_rec << "\n Lig:\t\t\t" << gpol_lig << endl;
		cout << "\nTotal:\t\t\t" << total_score << endl << endl;
*/
	}

	return true;
}

/************************************************/
string GB_Hawkins::output_score_summary() {
  string	text;
  char		line[1000];

  if(use_score) {
	  sprintf(line, "########## GB/SA Score:%20f\n##########%14s%20f\n##########%14s%20f\n##########%14s%20f\n##########\n", current_score,"vdw:", vdw_component, "es:", es_component, "GB:", (gpol_cmp-(gpol_rec+gpol_lig)));
	  text = line;

	  return text;
  } else
	  return "";
}

/************************************************/
void GB_Hawkins::prepare_receptor() {
	int		i, j;
	float	distance2;

	// Hawkins, Cramer & Truhlar variables
	float	Rkkp, ialphak, Pk, Pkp, sum;

	// Gpol variables
	float	epsout, gfac1, totgpol, gpself, gpshld, alphaij, alphaij2, Dij, fgb, engshld, engself;

	gpol_rec = 0.0;

	if(use_score) {

		if(rec_dist_assigned)
			delete [] rec_dist_mat;

		rec_dist_assigned = true;
		rec_dist_mat = new float[receptor.num_atoms * receptor.num_atoms];

		if(brad_rec_assigned)
			delete [] born_radius_rec;

		brad_rec_assigned = true;
		born_radius_rec = new float[receptor.num_atoms];

		for(i=0;i<receptor.num_atoms;i++) {

			sum = 0.0;
			Pk = receptor.gb_hawkins_radius[i] - gb_offset;

			for(j=0;j<receptor.num_atoms;j++) {

				if((i != j)&&(receptor.atom_active_flags[i])&&(receptor.atom_active_flags[j])) {

					// calc atom pair distance & init variables
					distance2 = pow((receptor.x[i]-receptor.x[j]),2) +
								pow((receptor.y[i]-receptor.y[j]),2) +
								pow((receptor.z[i]-receptor.z[j]),2) ;

					rec_dist_mat[i*receptor.num_atoms + j] = distance2;

					Rkkp = sqrt(distance2);
					Pkp = receptor.gb_hawkins_scale[j] * (receptor.gb_hawkins_radius[j] - gb_offset);

					sum = sum + born_radii_calc(Rkkp, Pkp, Pk);
				}

			} // end second atom loop

			sum = sum/2.0;
			ialphak = (1.0/Pk)-sum;
			born_radius_rec[i] = (1.0/ialphak);

			if(born_radius_rec[i] <= 0.0) {
				born_radius_rec[i] = 1000000.0;
				cout << "Born Radius Error: radius set to 1000000.0 for receptor atom " << i << endl;
			}

		} // end first atom loop


		// calculate the Solvent Polarization Energy (gpol)
		epsout = 80.0;
		gfac1 = -166.0*(1.0-(1.0/epsout));
		totgpol = 0.0;
		gpself = 0.0;
		gpshld = 0.0;

		for(i=0;i<receptor.num_atoms;i++) {

			distance2 = 0.0;

			for(j=i+1;j<receptor.num_atoms;j++) {

				if((i!=j)&&(receptor.atom_active_flags[i])&&(receptor.atom_active_flags[j])) {

					// calc atom pair distance & init variables
//					distance2 = pow((receptor.x[i]-receptor.x[j]),2) +
//							    pow((receptor.y[i]-receptor.y[j]),2) +
//							    pow((receptor.z[i]-receptor.z[j]),2);

					distance2 = rec_dist_mat[i*receptor.num_active_atoms + j];

					alphaij2 = born_radius_rec[i]*born_radius_rec[j];
					alphaij  = sqrt(alphaij2);

					Dij = distance2/(4.0*alphaij2);
					fgb = sqrt(distance2 + (alphaij2*exp(-Dij)));

					engshld = (gfac1 * receptor.charges[i] * receptor.charges[j] )/fgb;
					gpshld = gpshld + engshld;

				} // end calculation inner loop

			} // end second atom loop

			engself = (gfac1 * receptor.charges[i] * receptor.charges[i]) / born_radius_rec[i];
			gpself = gpself +  engself;

		} // end first atom loop

		gpshld = gpshld*2;
		totgpol = gpself + gpshld;
		gpol_rec = totgpol;

	} // end if use score

	cout << endl << "Receptor (Self, Shld, Total):\t" << gpself << "\t" << gpshld << "\t" << gpol_rec << endl;
}

/************************************************/
void GB_Hawkins::prepare_ligand(DOCKMol &ligand) {
	int		i, j;
	float	distance, distance2;

	// Hawkins, Cramer & Truhlar variables
	float	Rkkp, ialphak, Pk, Pkp, sum;

	// Gpol variables
	float	epsout, gfac1, totgpol, gpself, gpshld, alphaij, alphaij2, Dij, fgb, engshld, engself;

	gpol_lig = 0.0;

	if(use_score) {

		if(brad_lig_assigned)
			delete [] born_radius_lig;

		brad_lig_assigned = true;
		born_radius_lig = new float[ligand.num_atoms];

		for(i=0;i<ligand.num_atoms;i++) {

			sum = 0.0;
			Pk = ligand.gb_hawkins_radius[i] - gb_offset;

			for(j=0;j<ligand.num_atoms;j++) {

				if((i != j)&&(ligand.atom_active_flags[i])&&(ligand.atom_active_flags[j])) {

					// calc atom pair distance & init variables
					distance = sqrt( pow((ligand.x[i]-ligand.x[j]),2) +
									 pow((ligand.y[i]-ligand.y[j]),2) +
									 pow((ligand.z[i]-ligand.z[j]),2) );

					Rkkp = distance;
					Pkp = ligand.gb_hawkins_scale[j] * (ligand.gb_hawkins_radius[j] - gb_offset);

					sum = sum + born_radii_calc(Rkkp, Pkp, Pk);

				}

			} // end second atom loop

			sum = sum/2.0;
			ialphak = (1.0/Pk)-sum;
			born_radius_lig[i] = (1.0/ialphak);

			if(born_radius_lig[i] <= 0.0) {
				born_radius_lig[i] = 1000000.0;
				cout << "Born Radius Error: radius set to 1000000.0 for ligand atom " << i << endl;
			}

		} // end first atom loop

		// calculate the Solvent Polarization Energy (gpol)
		epsout = 80.0;
		gfac1 = -166.0*(1.0-(1.0/epsout));
		totgpol = 0.0;
		gpself = 0.0;
		gpshld = 0.0;

		for(i=0;i<ligand.num_atoms;i++) {

			distance2 = 0.0;

			for(j=i+1;j<ligand.num_atoms;j++) {

				if((i!=j)&&(ligand.atom_active_flags[i])&&(ligand.atom_active_flags[j])) {

					// calc atom pair distance & init variables
					distance2 = pow((ligand.x[i]-ligand.x[j]),2) +
							    pow((ligand.y[i]-ligand.y[j]),2) +
							    pow((ligand.z[i]-ligand.z[j]),2);

					alphaij2 = born_radius_lig[i]*born_radius_lig[j];
					alphaij  = sqrt(alphaij2);

					Dij = distance2/(4.0*alphaij2);
					fgb = sqrt(distance2 + (alphaij2*exp(-Dij)));

					engshld = (gfac1 * ligand.charges[i] * ligand.charges[j] )/fgb;
					gpshld = gpshld + engshld;

				} // end calculation inner loop

			} // end second atom loop

			engself = (gfac1 * ligand.charges[i] * ligand.charges[i]) / born_radius_lig[i];
			gpself = gpself +  engself;

		} // end first atom loop

		gpshld = gpshld*2;
		totgpol = gpself + gpshld;
		gpol_lig = totgpol;

	} // end if use score

	cout << endl << "Ligand   (Self, Shld, Total):\t" << gpself << "\t" << gpshld << "\t" << gpol_lig << endl;
}

/************************************************/
void GB_Hawkins::prepare_complex(DOCKMol &ligand) {
	int		i, j;
	float	distance, distance2;

	// Hawkins, Cramer & Truhlar variables
	float	Rkkp, ialphak, Pk, Pkp, sum;

	// Gpol variables
	float	epsout, gfac1, totgpol, gpself, gpshld, alphaij, alphaij2, Dij, fgb, engshld, engself;

	gpol_cmp = 0.0;

	if(use_score) {

		if(brad_cmp_assigned)
			delete [] born_radius_cmp;

		brad_cmp_assigned = true;
		born_radius_cmp = new float[ligand.num_atoms + receptor.num_atoms];

		vdw_component = es_component = 0.0;

		// loop over receptor atoms and calc the born radii
		for(i=0;i<receptor.num_atoms;i++) {

			sum = 0.0;
			Pk = receptor.gb_hawkins_radius[i] - gb_offset;

			// loop over receptor atoms
			for(j=0;j<receptor.num_atoms;j++) {

				if((i != j)&&(receptor.atom_active_flags[i])&&(receptor.atom_active_flags[j])) {

					// calc atom pair distance & init variables
//					distance = sqrt( pow((receptor.x[i]-receptor.x[j]),2) +
//									 pow((receptor.y[i]-receptor.y[j]),2) +
//									 pow((receptor.z[i]-receptor.z[j]),2) );

					distance = sqrt(rec_dist_mat[i*receptor.num_atoms + j]);

					Rkkp = distance;
					Pkp = receptor.gb_hawkins_scale[j] * (receptor.gb_hawkins_radius[j] - gb_offset);

					sum = sum + born_radii_calc(Rkkp, Pkp, Pk);

				}

			} // end second atom loop

			// loop over ligand atoms
			for(j=0;j<ligand.num_atoms;j++) {

				if((receptor.atom_active_flags[i])&&(ligand.atom_active_flags[j])) {

					// calc atom pair distance & init variables
					distance2 = pow((receptor.x[i]-ligand.x[j]),2) +
								pow((receptor.y[i]-ligand.y[j]),2) +
								pow((receptor.z[i]-ligand.z[j]),2) ;

					distance = sqrt(distance2);

					vdw_component += (((vdwA[receptor.amber_at_id[i]] * vdwA[ligand.amber_at_id[j]])/pow(distance, vdw_rep_exp)) - ((vdwB[receptor.amber_at_id[i]] * vdwB[ligand.amber_at_id[j]])/pow(distance, vdw_att_exp)));
					es_component  += ((332 * receptor.charges[i] * ligand.charges[j])/(pow(distance,2)));

					Rkkp = distance;
					Pkp = ligand.gb_hawkins_scale[j] * (ligand.gb_hawkins_radius[j] - gb_offset);

					sum = sum + born_radii_calc(Rkkp, Pkp, Pk);

				}

			} // end second atom loop

			sum = sum/2.0;
			ialphak = (1.0/Pk)-sum;
			born_radius_cmp[i] = (1.0/ialphak);

			if(born_radius_cmp[i] <= 0.0) {
				born_radius_cmp[i] = 1000000.0;
				cout << "Born Radius Error: radius set to 1000000.0 for complex (receptor) atom " << i << endl;
			}

		} // end first atom loop

		// loop over ligand atoms and calc the born radii
		for(i=0;i<ligand.num_atoms;i++) {

			sum = 0.0;
			Pk = ligand.gb_hawkins_radius[i] - gb_offset;

			// loop over receptor atoms
			for(j=0;j<receptor.num_atoms;j++) {

				if((ligand.atom_active_flags[i])&&(receptor.atom_active_flags[j])) {

					// calc atom pair distance & init variables
					distance = sqrt( pow((ligand.x[i]-receptor.x[j]),2) +
									 pow((ligand.y[i]-receptor.y[j]),2) +
									 pow((ligand.z[i]-receptor.z[j]),2) );

					Rkkp = distance;
					Pkp = receptor.gb_hawkins_scale[j] * (receptor.gb_hawkins_radius[j] - gb_offset);

					sum = sum + born_radii_calc(Rkkp, Pkp, Pk);

				}

			} // end second atom loop

			// loop over ligand atoms
			for(j=0;j<ligand.num_atoms;j++) {

				if((i!=j)&&(ligand.atom_active_flags[i])&&(ligand.atom_active_flags[j])) {

					// calc atom pair distance & init variables
					distance = sqrt( pow((ligand.x[i]-ligand.x[j]),2) +
									 pow((ligand.y[i]-ligand.y[j]),2) +
									 pow((ligand.z[i]-ligand.z[j]),2) );

					Rkkp = distance;
					Pkp = ligand.gb_hawkins_scale[j] * (ligand.gb_hawkins_radius[j] - gb_offset);

					sum = sum + born_radii_calc(Rkkp, Pkp, Pk);

				}

			} // end second atom loop

			sum = sum/2.0;
			ialphak = (1.0/Pk)-sum;
			born_radius_cmp[receptor.num_atoms + i] = (1.0/ialphak);

			if(born_radius_cmp[receptor.num_atoms + i] <= 0.0) {
				born_radius_cmp[receptor.num_atoms + i] = 1000000.0;
				cout << "Born Radius Error: radius set to 1000000.0 for complex (ligand) atom " << i << endl;
			}

		} // end first atom loop

		/////////////////////////////////////////////////////////////////////////////////

		// calculate the Solvent Polarization Energy (gpol)
		epsout = 80.0;
		gfac1 = -166.0*(1.0-(1.0/epsout));
		totgpol = 0.0;
		gpself = 0.0;
		gpshld = 0.0;

		// loop over receptor atoms
		for(i=0;i<receptor.num_atoms;i++) {

			// loop over receptor atoms
			for(j=i+1;j<receptor.num_atoms;j++) { // i+1

				if((i!=j)&&(receptor.atom_active_flags[i])&&(receptor.atom_active_flags[j])) {

					// calc atom pair distance & init variables
//					distance2 = pow((receptor.x[i]-receptor.x[j]),2) +
//							    pow((receptor.y[i]-receptor.y[j]),2) +
//							    pow((receptor.z[i]-receptor.z[j]),2);

					distance2 = rec_dist_mat[i*receptor.num_atoms + j];

					alphaij2 = born_radius_cmp[i]*born_radius_cmp[j];
					alphaij  = sqrt(alphaij2);

					Dij = distance2/(4.0*alphaij2);
					fgb = sqrt(distance2 + (alphaij2*exp(-Dij)));

					engshld = (gfac1 * receptor.charges[i] * receptor.charges[j] )/fgb;
					gpshld = gpshld + engshld;

				} // end calculation inner loop

			} // end receptor atom loop

			// loop over ligand atoms
			for(j=0;j<ligand.num_atoms;j++) {

				if((receptor.atom_active_flags[i])&&(ligand.atom_active_flags[j])) {

					// calc atom pair distance & init variables
					distance2 = pow((receptor.x[i]-ligand.x[j]),2) +
							    pow((receptor.y[i]-ligand.y[j]),2) +
							    pow((receptor.z[i]-ligand.z[j]),2);

					alphaij2 = born_radius_cmp[i]*born_radius_cmp[j];
					alphaij  = sqrt(alphaij2);

					Dij = distance2/(4.0*alphaij2);
					fgb = sqrt(distance2 + (alphaij2*exp(-Dij)));

					engshld = (gfac1 * receptor.charges[i] * ligand.charges[j] )/fgb;
					gpshld = gpshld + engshld;

				} // end calculation inner loop

			} // end ligand atom loop

			engself = (gfac1 * receptor.charges[i] * receptor.charges[i]) / born_radius_cmp[i];
			gpself = gpself +  engself;

		} // end receptor atom loop

		// loop over ligand atoms
		for(i=0;i<ligand.num_atoms;i++) {
/*
			// loop over receptor atoms
			for(j=0;j<receptor.num_atoms;j++) {

				if((ligand.atom_active_flags[i])&&(receptor.atom_active_flags[j])) {

					// calc atom pair distance & init variables
					distance2 = pow((ligand.x[i]-receptor.x[j]),2) +
							    pow((ligand.y[i]-receptor.y[j]),2) +
							    pow((ligand.z[i]-receptor.z[j]),2);

					alphaij = sqrt(born_radius_cmp[i]*born_radius_cmp[j]);
					alphaij2 = pow(alphaij, 2);

					Dij = distance2/(4.0*alphaij2);
					fgb = sqrt(distance2 + (alphaij2*exp(-Dij)));

					engshld = (gfac1 * ligand.charges[i] * receptor.charges[j] )/fgb;
					gpshld = gpshld + engshld;

				} // end calculation inner loop

			} // end receptor atom loop
*/
			// loop over ligand atoms
			for(j=i+1;j<ligand.num_atoms;j++) { // i+1

				if((i!=j)&&(ligand.atom_active_flags[i])&&(ligand.atom_active_flags[j])) {

					// calc atom pair distance & init variables
					distance2 = pow((ligand.x[i]-ligand.x[j]),2) +
							    pow((ligand.y[i]-ligand.y[j]),2) +
							    pow((ligand.z[i]-ligand.z[j]),2);

					alphaij2 = born_radius_cmp[i]*born_radius_cmp[j];
					alphaij  = sqrt(alphaij2);

					Dij = distance2/(4.0*alphaij2);
					fgb = sqrt(distance2 + (alphaij2*exp(-Dij)));

					engshld = (gfac1 * ligand.charges[i] * ligand.charges[j] )/fgb;
					gpshld = gpshld + engshld;

				} // end calculation inner loop

			} // end ligand atom loop

			engself = (gfac1 * ligand.charges[i] * ligand.charges[i]) / born_radius_cmp[receptor.num_atoms + i];
			gpself = gpself +  engself;

		} // end ligand atom loop

		gpshld = gpshld*2; // i+1
		totgpol = gpself + gpshld;
		gpol_cmp = totgpol;

	} // end if use score

	cout << endl << "Complex  (Self, Shld, Total):\t" << gpself << "\t" << gpshld << "\t" << gpol_cmp << endl;

}

/************************************************/
/************************************************/
/************************************************/
void Dreiding_Internal_Energy::input_parameters(Parameter_Reader &parm) {

	use_score = (parm.query_param("use_dreiding_int_nrg", "no", "yes no") == "yes") ? true : false;

	if(use_score) {
		atom_def_fname = parm.query_param("dreiding_atom_type_file", "dreiding_atom_types.defn");
		bond_def_fname = parm.query_param("dreiding_bond_type_file", "dreiding_bond_types.defn");
	}

}

/************************************************/
void Dreiding_Internal_Energy::initialize() {

	atom_types_allocated = false;
	bond_types_allocated = false;
	mol_types_allocated  = false;

	if(use_score) {
		read_atom_type_defs(atom_def_fname);
		read_bond_type_defs(bond_def_fname);
	}

}

/************************************************/
void Dreiding_Internal_Energy::read_atom_type_defs(string fname) {
	FILE					*ifp;
	char					line[1000];
	char					tmp_line[1000];
	string					tmp_string;
	DREIDING_ATOM_TYPE		tmp_type;
	int						index;

	// clear types if they already exist
	if(atom_types_allocated) {
		delete [] atom_types;
		num_atom_types = 0;
	}

	// open dreiding atom type .defn file
	ifp = fopen(fname.c_str(), "r");

	if(ifp == NULL) {
		cout << "\n\nCould not open " << fname << " for reading.  Program will terminate." << endl << endl;
		exit(0);
	}

	// count the number of types
	num_atom_types = 0;

	while (fgets (line, 1000, ifp) != NULL) {
		if (!strncmp (line, "name", 4))
			num_atom_types++;
	}

	rewind(ifp);

	// allocate type vector
	atom_types = new DREIDING_ATOM_TYPE[num_atom_types];
	atom_types_allocated = true;

	// read lines into atom type vector
	index = -1;

	while (fgets (line, 100, ifp) != NULL) {

		if (!strncmp (line, "name", 4)) { // read in name field
			index++;

			if (sscanf (line, "%*s %s", tmp_line) < 1) {
				cout << "Incomplete dreiding member declaration.\n";
				exit (0);
			}

			atom_types[index].name = tmp_line;

		} else if (!strncmp (line, "bond_radius", 11)) { // read in bond_radius field
			if (sscanf (line, "%*s %f", &atom_types[index].bond_radius) != 1) {
				cout << "Incomplete bond radius specification.\n";
				exit (0);
			}
		} else if (!strncmp (line, "bond_angle", 10)) { // read in bond angle field
			if (sscanf (line, "%*s %f", &atom_types[index].bond_angle) != 1) {
				cout << "Incomplete bond angle specification.\n";
				exit (0);
			}
		} else if (!strncmp (line, "hybrid", 6)) { // read in well depth field
			if (sscanf (line, "%*s %s", tmp_line) != 1) {
				cout << "Incomplete hybridization specification.\n";
				exit (0);
			}

			atom_types[index].hybrid = tmp_line;

		} else if (!strncmp (line, "definition", 10)) { // read in definition fields
			if (sscanf (line, "%*s %s", tmp_line) != 1) {
				cout << "Error assigning dreiding definitions.\n";
				exit (0);
			}

			tmp_string = tmp_line;
			atom_types[index].definitions.push_back(tmp_string);
		}

	} // end While

}

/************************************************/
void Dreiding_Internal_Energy::read_bond_type_defs(string fname) {
	FILE					*ifp;
	char					line[1000];
	char					tmp_line[1000];
	string					tmp_string;
	DREIDING_ATOM_TYPE		tmp_type;
	int						index;

	// clear types if they already exist
	if(bond_types_allocated) {
		delete [] bond_types;
		num_bond_types = 0;
	}

	// open dreiding atom type .defn file
	ifp = fopen(fname.c_str(), "r");

	if(ifp == NULL) {
		cout << "\n\nCould not open " << fname << " for reading.  Program will terminate." << endl << endl;
		exit(0);
	}

	// count the number of types
	num_bond_types = 0;

	while (fgets (line, 1000, ifp) != NULL) {
		if (!strncmp (line, "name", 4))
			num_bond_types++;
	}

	rewind(ifp);

	// allocate type vector
	bond_types = new DREIDING_BOND_TYPE[num_bond_types];
	bond_types_allocated = true;

	// read lines into atom type vector
	index = -1;

	while (fgets (line, 100, ifp) != NULL) {

		if (!strncmp (line, "name", 4)) { // read in name field
			index++;

			if (sscanf (line, "%*s %s", tmp_line) < 1) {
				cout << "Incomplete dreiding member declaration.\n";
				exit (0);
			}

			bond_types[index].name = tmp_line;

		} else if (!strncmp (line, "period", 6)) { // read in period field
			if (sscanf (line, "%*s %d", &bond_types[index].n) != 1) {
				cout << "Incomplete bond period specification.\n";
				exit (0);
			}
		} else if (!strncmp (line, "barrier", 7)) { // read in bond barrier field
			if (sscanf (line, "%*s %f", &bond_types[index].v) != 1) {
				cout << "Incomplete bond barrier specification.\n";
				exit (0);
			}
		} else if (!strncmp (line, "angle", 5)) { // read in bond angle field
			if (sscanf (line, "%*s %f", &bond_types[index].angle) != 1) {
				cout << "Incomplete bond angle specification.\n";
				exit (0);
			}
		} else if (!strncmp (line, "hybs1", 5)) { // read in hybridization fields

			bond_types[index].hybs1 = line;

		} else if (!strncmp (line, "hybs2", 5)) { // read in hybridization fields

			bond_types[index].hybs2 = line;

		} else if (!strncmp (line, "defs1", 5)) { // read in atom type def fields

			bond_types[index].defs1 = line;

		} else if (!strncmp (line, "defs2", 5)) { // read in atom type def fields

			bond_types[index].defs2 = line;

		} else if (!strncmp (line, "bond", 4)) { // read in atom type def fields

			bond_types[index].bond = line;

		}

	} // end While

}

/************************************************/
void Dreiding_Internal_Energy::assign_types_to_mol(DOCKMol &mol) {
	int			i, j, k;
	int			a1, a2;
	bool		type_matched;

	if(use_score) {

		// perceive torsions
		id_torsions(mol);

		// clear types if they've been assigned
		if(mol_types_allocated) {
			delete [] atoms;
			delete [] bonds;
		}

		atoms = new int[mol.num_atoms];
		bonds = new int[mol.num_bonds];

		// loop over and assign atom types
		for(i=0;i<mol.num_atoms;i++) {
			type_matched = false;

			for(j=0;j<num_atom_types;j++) {
				for(k=0;k<atom_types[j].definitions.size();k++) {
					if((atom_types[j].definitions[k] == "*") || (mol.atom_types[i] == atom_types[j].definitions[k])) {
						type_matched = true;
						atoms[i] = j;
					}
				}
			}

			if(!type_matched)
				cout << "Warning: Dreiding Atom type not found for atom " << i << endl;
		}

		// loop over and assign bond types
		for(i=0;i<mol.num_bonds;i++) {
			type_matched = false;

			a1 = mol.bonds_origin_atom[i];
			a2 = mol.bonds_target_atom[i];

			for(j=0;j<num_bond_types;j++) {

				// check if bond origin atom matches hybs1 in definition
				if((strstr(bond_types[j].hybs1.c_str(), "*")) || (strstr(bond_types[j].hybs1.c_str(), atom_types[atoms[a1]].hybrid.c_str()))) {

					if((strstr(bond_types[j].defs1.c_str(), "*")) || (strstr(mol.atom_types[a1].c_str(), bond_types[j].defs1.c_str()))) {

						if((strstr(bond_types[j].hybs2.c_str(), "*")) || (strstr(bond_types[j].hybs2.c_str(), atom_types[atoms[a2]].hybrid.c_str()))) {

							if((strstr(bond_types[j].defs2.c_str(), "*")) || (strstr(mol.atom_types[a2].c_str(), bond_types[j].defs2.c_str()))) {

								if( (strstr(bond_types[j].bond.c_str(), "*")) || (strstr(bond_types[j].bond.c_str(), mol.bond_types[i].c_str()))) {
									type_matched = true;
									bonds[i] = j;
								}

							}

						}

					}

				}
				/////////////////////////////////////////////////////////

				// check if bond origin atom matches hybs1 in definition
				if((strstr(bond_types[j].hybs1.c_str(), "*")) || (strstr(bond_types[j].hybs1.c_str(), atom_types[atoms[a2]].hybrid.c_str()))) {

					if((strstr(bond_types[j].defs1.c_str(), "*")) || (strstr(mol.atom_types[a2].c_str(), bond_types[j].defs1.c_str()))) {

						if((strstr(bond_types[j].hybs2.c_str(), "*")) || (strstr(bond_types[j].hybs2.c_str(), atom_types[atoms[a1]].hybrid.c_str()))) {

							if((strstr(bond_types[j].hybs2.c_str(), "*")) || (strstr(bond_types[j].hybs2.c_str(), atom_types[atoms[a1]].hybrid.c_str()))) {

								if( (strstr(bond_types[j].bond.c_str(), "*")) || (strstr(bond_types[j].bond.c_str(), mol.bond_types[i].c_str()))) {
									type_matched = true;
									bonds[i] = j;
								}

							}

						}

					}

				}
				/////////////////////////////////////////////////////////

			}

			if(!type_matched)
				cout << "Warning: Dreiding bond type not found for bond " << i << endl;

		}

	}

}

/************************************************/
float Dreiding_Internal_Energy::calc_bond_torsion_term(DOCKMol &mol, int a1, int a2, int a3, int a4) {
	int		bond_num;
	float	bond_angle, energy;

	bond_num   = mol.get_bond(a2, a3);
	bond_angle = mol.get_torsion(a1, a2, a3, a4);

	energy = (0.5) * bond_types[bonds[bond_num]].v * (1 - cos(bond_types[bonds[bond_num]].n * (bond_angle - bond_types[bonds[bond_num]].angle)));

	return energy;
}

/************************************************/
float Dreiding_Internal_Energy::compute_internal_energy(DOCKMol &mol) {
	int		i;
	float	energy;

	energy = 0.0;

	if(use_score) {

		for(i=0;i<torsions.size();i++) {
			if((mol.bond_active_flags[torsions[i].bond_num])&&(mol.bond_is_rotor(torsions[i].bond_num))) {
				energy += calc_bond_torsion_term(mol, torsions[i].atom1, torsions[i].atom2, torsions[i].atom3, torsions[i].atom4);
			}
		}

	}

	return energy;
}

/******************************************************/
void Dreiding_Internal_Energy::id_torsions(DOCKMol &mol) {
	int				i, j, max_central, central;
	int				a1, a2, a3, a4;
	int				nbr;
	vector<int>		nbrs;
	TORSION			tmp_torsion;

	BREADTH_SEARCH	bfs;

	torsions.clear();

	//loop over bonds- add flex bonds to torsion list
	for(i=0;i<mol.num_bonds;i++) {
		if(mol.bond_active_flags[i]) {
			if(mol.bond_is_rotor(i)) {
				torsions.push_back(tmp_torsion);
				torsions[torsions.size()-1].bond_num = i;
			}
		}
	}

	// ID the inter-segment rot-bonds
	for(i=0;i<torsions.size();i++) {

		a2 = mol.bonds_origin_atom[torsions[i].bond_num];
		a3 = mol.bonds_target_atom[torsions[i].bond_num];

		max_central = -1;
//		nbrs = mol.get_atom_neighbors(a2);
		nbrs = mol.neighbor_list[a2];

		for (j=0;j<nbrs.size();j++) {
			nbr = nbrs[j];
			if(nbr != a3) {
				central = bfs.get_search_radius(mol, nbr, a2);
				if(central > max_central) {
					a1 = nbr;
					max_central = central;
				}
			}
		}

		max_central = -1;
//		nbrs = mol.get_atom_neighbors(a3);
		nbrs = mol.neighbor_list[a3];

		for (j=0;j<nbrs.size();j++) {
			nbr = nbrs[j];
			if(nbr != a2) {
				central = bfs.get_search_radius(mol, nbr, a3);
				if(central > max_central) {
					a4 = nbr;
					max_central = central;
				}
			}
		}

		torsions[i].atom1 = a1;
		torsions[i].atom2 = a2;
		torsions[i].atom3 = a3;
		torsions[i].atom4 = a4;
	}

}

