using namespace std;

#define VDWOFF -0.09
#define MAX_ATOM_REC 1000
#define MAX_ATOM_LIG 200

#define SQR(x) ((x)*(x))
#define MAX(x, y) ((x) > (y) ? (x) : (y))

/*****************************************************************/
class Bump_Filter : public Base_Score {

public:

	FILE			*grid_in;
	string			file_prefix;

	unsigned char	*bump;
	int				bump_filter;
	int				bump_max;

	void			input_parameters(Parameter_Reader &parm);
	void			initialize();
//	void			add_score_to_mol(DOCKMol &, string, float);

	void			read_bump_grid();
	bool			check_bumps(DOCKMol &mol, bool bypass);
	int				get_bump_score(DOCKMol &mol);

};

/*****************************************************************/
class Energy_Score : public Base_Score {

public:

	FILE			*grid_in;
	string			file_prefix;

	int				energy_score;
	float			*avdw;
	float			*bvdw;
	float			*es;
	int				att_exp;
	int				rep_exp;
	float			vdw_scale;
	float			es_scale;
	float			*vdwA;
	float			*vdwB;
	bool			vdwA_allocated;
	bool			vdwB_allocated;
	char			atom_model;
	float			vdw_component;
	float			es_component;

	void			input_parameters(Parameter_Reader &parm, bool &primary_score, bool &secondary_score);
	void			initialize(AMBER_TYPER &);

	void			read_nrg_grid();
	bool			compute_score(DOCKMol &mol);
	void			init_vdw_energy(AMBER_TYPER &);
	string			output_score_summary(float);

};

/*****************************************************************/
class Continuous_Energy_Score : public Base_Score {

public:

	string			receptor_filename;
	DOCKMol			receptor;

	int				att_exp;
	int				rep_exp;
	float			vdw_scale;
	float			es_scale;
	float			diel_screen;

	float			*vdwA;
	float			*vdwB;

	bool			vdwA_allocated;
	bool			vdwB_allocated;

	float			vdw_component;
	float			es_component;

	void			input_parameters(Parameter_Reader &parm, bool &primary_score, bool &secondary_score);
	void			initialize(AMBER_TYPER &);
	void			init_vdw_energy(AMBER_TYPER &typer);
	bool			compute_score(DOCKMol &);
	string			output_score_summary(float);

};

/*****************************************************************/
class Contact_Score : public Base_Score {

public:

	FILE			*grid_in;
	char			fname[100];
	char			bump_fname[100];
	string			file_prefix;

	int				contact_score;
	float			cutoff_distance;
	float			clash_overlap;
	float			clash_penalty;
	short int		*cnt;
	unsigned char	*bump;

	void			input_parameters(Parameter_Reader &parm, bool &primary_score, bool &secondary_score);
	void			initialize();

	void			read_cnt_grid();
	bool			compute_score(DOCKMol &mol);
	string			output_score_summary(float);

};

/*****************************************************************/
class Chemical_Score : public Base_Score {

public:

	FILE			*grid_in;
	char			fname[100];
	char			bump_fname[100];
	string			file_prefix;

	int				grid_num;

	void			input_parameters(Parameter_Reader &parm);
	void			initialize();

	void			read_chm_grid();

};

/*****************************************************************/
class COORDS {
public:
	float	crd[3];
};

typedef vector<COORDS> CRDVec;

class INT2000 {
public:
	int		vals[2000];
};

class GB_Pairwise : public Base_Score {

public:

	// GB Variable Section ///////////////////////////////
	// Receptor data
	int				gb_rec_total_atoms;
	float			gb_rec_solv_rec;
	vector<COORDS>	gb_rec_coords;
	FLOATVec		gb_rec_radius_eff;
	FLOATVec		gb_rec_charge;
	FLOATVec		gb_rec_inv_a_rec;
	FLOATVec		gb_rec_inv_a;
	FLOATVec		gb_rec_receptor_screen;

	// Ligand data
	vector<COORDS>	gb_lig_coords;
	FLOATVec		gb_lig_radius_eff;
	FLOATVec		gb_lig_sum_lig;
	FLOATVec		gb_lig_inv_a;
	INTVec			gb_lig_flag;

	// Grid data
	float			gb_grid_origin[3];
	float			gb_grid_spacing;
	float			gb_grid_grdcut;
	float			gb_grid_grdcuto;
	float			gb_grid_cutsq;
	float			gb_grid_f_scale;
	int				gb_grid_span[3];
	int				gb_grid_size;

	// control variables
	bool			gb_initialized;

	// SA Variable Section //////////////////////////////
	// Receptor data
	int				sa_rec_total_atoms;
	float			sa_rec_solv_rec;
	vector<COORDS>	sa_rec_coords;
	FLOATVec		sa_rec_radius_eff;
	INTVec			sa_rec_vdwn_rec;
	INTVec			sa_rec_nsphgrid;
	INTVec			sa_rec_nhp;

	int				sa_rec_nsas;
	int				sa_rec_nsas_hp;
	int				sa_rec_nsas_pol;

	// Ligand data
	vector<INT2000>	sa_grid_mark_sas_lig;
	vector<COORDS>	sa_lig_coords;
	FLOATVec		sa_lig_radius_eff;

	// Grid data
	float			sa_grid_origin[3];
	float			sa_grid_spacing;
	float			sa_grid_grdcut;
	float			sa_grid_grdcuto;
	float			sa_grid_cutsq;
	float			sa_grid_f_scale;
	int				sa_grid_span[3];
	int				sa_grid_size;
	int				sa_grid_nvtyp;

	float			sa_grid_r_probe;
	float			sa_grid_sspacing;
	float			sa_grid_r2_cutoff;
	INTVecVec		sa_grid_mark_sas;
	vector<CRDVec>	sa_grid_sphgrid_crd;

	// VDW calculation section
	Energy_Score	vdw_score;
	string			vdw_grid_prefix;

	// General Variablea
	int				use_gbsa;
	string			gb_grid_prefix;
	string			sa_grid_prefix;
	string			screen_file;
	float			solvent_dielectric;
	int				verbosity_level;
	float			vdw_component;
	float			gb_component;
	float			sa_component;
	float			del_SAS_hp;

	// Functions ////////////////////////////////////////////
	void			input_parameters(Parameter_Reader &parm, bool &primary_score, bool &secondary_score);
	void			initialize(AMBER_TYPER &typer, int argc, char **argv);
	bool			compute_score(DOCKMol &mol);
	float			return_current_score() { return current_score; };
	string			output_score_summary(float);

	void			read_gb_receptor_grid();
	float			get_gb_solvation_score(DOCKMol &mol);

	void			read_sa_receptor_grid();
	float			get_sa_solvation_score(DOCKMol &mol);
};

/*************************************************************/
class GB_Hawkins : public Base_Score {

public:

	// parameters
	string					rec_filename;
	int						vdw_att_exp;
	int						vdw_rep_exp;
	float					gb_offset;

	// internal data structures for GB
	DOCKMol					receptor;
	bool					brad_rec_assigned;
	bool					brad_lig_assigned;
	bool					brad_cmp_assigned;
	float					*born_radius_rec;
	float					*born_radius_lig;
	float					*born_radius_cmp;
	float					gpol_rec;
	float					gpol_lig;
	float					gpol_cmp;

	bool					rec_dist_assigned;
	float					*rec_dist_mat;

	// internal data for VDW & ES score
	float					*vdwA;
	float					*vdwB;
	bool					vdwA_assigned;
	bool					vdwB_assigned;
	float					vdw_component;
	float					es_component;
	float					total_score;

	// Energy Scoring
//	Continuous_Energy_Score	c_nrg;

	void					input_parameters(Parameter_Reader &parm, bool &primary_score, bool &secondary_score);
	void					initialize(AMBER_TYPER &);
	void					init_vdw_energy(AMBER_TYPER &);

	void					prepare_receptor();
	void					prepare_ligand(DOCKMol &);
	void					prepare_complex(DOCKMol &);

	bool					compute_score(DOCKMol &mol);
	float					return_current_score() { return current_score; };
	string					output_score_summary();

};

/************************************************/
/************************************************/
/************************************************/
class DREIDING_ATOM_TYPE {

public:
	string				name;
	float				bond_radius;
	float				bond_angle;
	string				hybrid;
	vector<string>		definitions;

};

/************************************************/
class DREIDING_BOND_TYPE {

public:
	string				name;
	string				hybs1, hybs2;
	string				defs1, defs2;
	string				bond;

	int					n;
	float				v;
	float				angle;

};

/************************************************/
class Dreiding_Internal_Energy {

public:

	bool							use_score;
	string							atom_def_fname;
	string							bond_def_fname;

	bool							atom_types_allocated;
	int								num_atom_types;
	DREIDING_ATOM_TYPE				*atom_types;

	bool							bond_types_allocated;
	int								num_bond_types;
	DREIDING_BOND_TYPE				*bond_types;

	// molecule atom & bond type arrays
	bool							mol_types_allocated;
	int								*atoms;
	int								*bonds;

	vector<TORSION>					torsions;

	// general functions
	void							input_parameters(Parameter_Reader &parm);
	void							initialize();

	// atom/bond type functions
	void							read_atom_type_defs(string);
	void							read_bond_type_defs(string);
	void							assign_types_to_mol(DOCKMol &);
	float							calc_bond_torsion_term(DOCKMol &, int, int, int, int);
	float							compute_internal_energy(DOCKMol &mol);
	void							id_torsions(DOCKMol &);

};

