using namespace std;

/********************************************************************/
class Simplex_Minimizer {

public:

	vector<TORSION>	torsions;
	INTVec			torsion_scale_factors;
	INTVec			bond_vectors;

	// state variables
	int				current_cycle;
	int				random_seed;

	// simplex parameters
	bool			advanced_min_params;
//	int				max_iterations;////
	int				max_cycles;////
	float			score_converge;////
	float			cycle_converge;////
	float			trans_step_size;////
	float			rot_step_size;  ////
	float			tors_step_size; ////

	// anchor minimization parameters
	int				anchor_min_max_iterations;
	int				anchor_min_max_cycles;
	float			anchor_min_score_converge;
	float			anchor_min_cycle_converge;
	float			anchor_min_trans_step_size;
	float			anchor_min_rot_step_size;
	float			anchor_min_tors_step_size;

	// flex search minimization parameters
	int				flex_min_max_iterations;
	int				flex_min_max_cycles;
	float			flex_min_score_converge;
	float			flex_min_cycle_converge;
	float			flex_min_trans_step_size;
	float			flex_min_rot_step_size;
	float			flex_min_tors_step_size;

	// final minimization parameters
	int				final_min_max_iterations;
	int				final_min_max_cycles;
	float			final_min_score_converge;
	float			final_min_cycle_converge;
	float			final_min_trans_step_size;
	float			final_min_rot_step_size;
	float			final_min_tors_step_size;

	// control parameters
	bool			use_min_rigid_anchor;
	bool			use_min_flex_growth;
	bool			use_min_final_pose;

/***
	// monte carlo preminimizer parameters
	bool			use_mc_premin;
	bool			mc_premin_override;
	bool			use_mc_postmin;
	float			r_value;
	int				fail_threshold;
	int				max_steps;
***/

	// functions
	void			input_parameters(Parameter_Reader &parm);
	void			initialize();
	void			minimize(Base_Score &, DOCKMol &, FLOATVec &, int, float, int, float, float, float, float);
	float			simplex_minimize(Base_Score &, DOCKMol &, FLOATVec &, int, float, float, float, float);
	bool			simplex_score(Base_Score &, DOCKMol &, DOCKMol &, FLOATVec &, float, float, float);
	void			scale_simplex_vector(FLOATVec &, FLOATVec &, float, float, float);
	void			vector_to_dockmol(DOCKMol &, FLOATVec &);
	void			minimize_rigid_anchor(DOCKMol &, Master_Score &);
	void			minimize_flexible_growth(DOCKMol &, Master_Score &, INTVec &);
	void			minimize_flexible_growth(DOCKMol &, Master_Score &);
	void			minimize_final_pose(DOCKMol &, Master_Score &);
	void			id_torsions(DOCKMol &, FLOATVec &);

};

